"""Set RACF Options Administration."""

from typing import List, Tuple, Union

from pyracf.common.security_admin import SecurityAdmin

from .setropts_requset import SetroptsRequest


class SetroptsAdmin(SecurityAdmin):
    """Set RACF Options Administration."""

    def __init__(
        self,
        debug: bool = False,
        generate_requests_only: bool = False,
        update_existing_segment_traits: Union[dict, None] = None,
        replace_existing_segment_traits: Union[dict, None] = None,
        additional_secret_traits: Union[List[str], None] = None,
    ) -> None:
        self._valid_segment_traits = {
            "base": {
                "base:active_classes": "racf:classact",
                "base:add_creator": "racf:addcreat",
                "base:automatic_data_set_protection": "racf:adsp",
                "base:application_logon_auditing": "racf:applaudt",
                "base:audit_classes": "racf:audit",
                "base:uncataloged_data_set_access": "racf:catdsns",
                "base:log_racf_command_violations": "racf:cmdviol",
                "base:security_label_compatibility_mode": "racf:compmode",
                "base:enhanced_generic_naming": "racf:egn",
                "base:erase_data_sets_on_delete": "racf:erase",
                "base:erase_data_sets_on_delete_all": "racf:eraseall",
                "base:erase_data_sets_on_delete_security_level": "racf:erasesec",
                "base:generic_command_classes": "racf:gencmd",
                "base:generic_profile_checking_classes": "racf:generic",
                "base:generic_profile_sharing_classes": "racf:genlist",
                "base:generic_owner": "racf:genowner",
                "base:global_access_classes": "racf:global",
                "base:list_of_groups_access_checking": "racf:grplist",
                "base:password_history": "racf:history",
                "base:revoke_inactive_userids_interval": "racf:inactive",
                "base:record_user_verification_statistics": "racf:initstat",
                "base:max_password_change_interval": "racf:interval",
                "base:jes_batch": "racf:jesbatch",
                "base:jes_early_verification": "racf:jesearly",
                "base:jes_network_user": "racf:jesnje",
                "base:jes_undefined_user": "racf:jesundef",
                "base:jes_execution_batch_monitoring": "racf:jesxbm",
                "base:kerberos_encryption_level": "racf:kerblvl",
                "base:list": "racf:list",
                "base:audit_log_always_classes": "racf:logalwys",
                "base:audit_log_default_classes": "racf:logdeflt",
                "base:audit_log_failure_classes": "racf:logfail",
                "base:audit_log_never_classes": "racf:lognever",
                "base:audit_log_success_classes": "racf:logsucc",
                "base:min_password_change_interval": "racf:minchang",
                "base:mixed_case_password_support": "racf:mixdcase",
                "base:multi_level_security_address_space": "racf:mlactive",
                "base:multi_level_security_file_system": "racf:mlfs",
                "base:multi_level_security_interprocess": "racf:mlipc",
                "base:multi_level_security_file_names": "racf:mlnames",
                "base:multi_level_security_logon": "racf:mlquiet",
                "base:multi_level_security_declassification": "racf:mls",
                "base:multi_level_security_label_alteration": "racf:mlstable",
                "base:profile_modelling": "racf:model",
                "base:profile_modelling_generation_data_group": "racf:modgdg",
                "base:profile_modelling_group": "racf:modgroup",
                "base:profile_modelling_user": "racf:moduser",
                "base:log_operator_actions": "racf:operaudt",
                "base:passphrase_change_interval": "racf:phrint",
                "base:data_set_single_level_name_prefix_protection": "racf:prefix",
                "base:primary_language": "racf:primlang",
                "base:protect_all_data_sets": "racf:protall",
                "base:password_encryption_algorithm": "racf:pwdalg",
                "base:special_character_password_support": "racf:pwdspec",
                "base:raclist": "racf:raclist",
                "base:log_real_data_set_name": "racf:realdsn",
                "base:refresh": "racf:refresh",
                "base:tape_data_set_security_retention_period": "racf:retpd",
                "base:max_incorrect_password_attempts": "racf:revoke",
                "base:password_rules": "racf:rules",
                "base:password_rule_1": "racf:rule1",
                "base:password_rule_2": "racf:rule2",
                "base:password_rule_3": "racf:rule3",
                "base:password_rule_4": "racf:rule4",
                "base:password_rule_5": "racf:rule5",
                "base:password_rule_6": "racf:rule6",
                "base:password_rule_7": "racf:rule7",
                "base:password_rule_8": "racf:rule8",
                "base:rvary_switch_password": "racf:rvarswpw",
                "base:rvary_status_password": "racf:rvarstpw",
                "base:log_commands_issuesd_by_special_users": "racf:saudit",
                "base:security_label_control": "racf:seclabct",
                "base:secondary_language": "racf:seclang",
                "base:session_key_verification_interval": "racf:sessint",
                "base:security_label_auditing": "racf:slabaudt",
                "base:security_label_system": "racf:slbysys",
                "base:security_level_auditing": "racf:slevaudt",
                "base:statistics_classes": "racf:classtat",
                "base:tape_data_set_protection": "racf:tapedsn",
                "base:terminal_universal_access": "racf:terminal",
                "base:password_expiration_warning": "racf:warning",
                "base:program_control": "racf:whenprog",
            }
        }
        super().__init__(
            "systemSettings",
            debug=debug,
            generate_requests_only=generate_requests_only,
            update_existing_segment_traits=update_existing_segment_traits,
            replace_existing_segment_traits=replace_existing_segment_traits,
            additional_secret_traits=additional_secret_traits,
        )

    # ============================================================================
    # Password Rules
    # ============================================================================
    def get_password_rules(self) -> Union[dict, bytes]:
        """Get RACF password rules."""
        profile = self.list_racf_options(options_only=True)
        return self._get_field(profile, "passwordProcessingOptions", "syntaxRules")

    # ============================================================================
    # Raclist Refresh
    # ============================================================================
    def refresh_raclist(self, class_names: Union[str, List[str]]) -> Union[dict, bytes]:
        """Refresh raclist."""
        result = self.alter(options={"base:raclist": class_names, "base:refresh": True})
        return self._to_steps(result)

    # ============================================================================
    # Get attributes
    # ============================================================================
    def get_class_attributes(self, class_name: str) -> Union[list, bytes]:
        """Get RACF get attributes."""
        profile = self.list_racf_options(options_only=True)
        if not isinstance(profile, dict):
            # Allows this function to work with "self._generate_requests_only" mode.
            return profile
        return [
            class_type
            for class_type in profile["classes"].keys()
            if class_name.lower() in profile["classes"][class_type]
        ]

    # ============================================================================
    # Audit Classes
    # ============================================================================
    def add_audit_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Add class(es) to list of classes that RACF performs auditing for."""
        result = self.alter(options={"base:audit_classes": class_names})
        return self._to_steps(result)

    def remove_audit_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Remove class(es) from the list of classes that RACF performs auditing for."""
        result = self.alter(options={"delete:base:audit_classes": class_names})
        return self._to_steps(result)

    # ============================================================================
    # Active Classes
    # ============================================================================
    def add_active_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Add class(es) to the list of classes that RACF performs access authorization checking for.
        """
        result = self.alter(options={"base:active_classes": class_names})
        return self._to_steps(result)

    def remove_active_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Remove class(es) from the list of classes that
        RACF performs access authorization checking for.
        """
        result = self.alter(options={"delete:base:active_classes": class_names})
        return self._to_steps(result)

    # ============================================================================
    # Statistics Classes
    # ============================================================================
    def add_statistics_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Add class(es) to the list of classes that RACF collects statistics for."""
        result = self.alter(options={"base:statistics_classes": class_names})
        return self._to_steps(result)

    def remove_statistics_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Remove class(es) from the list of classes that RACF collects statistics for."""
        result = self.alter(options={"delete:base:statistics_classes": class_names})
        return self._to_steps(result)

    # ============================================================================
    # Generic Command Processing Classes
    # ============================================================================
    def add_generic_command_processing_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Add class(es) to the list of classes that have
        generic profile command processing enabled.
        """
        result = self.alter(options={"base:generic_command_classes": class_names})
        return self._to_steps(result)

    def remove_generic_command_processing_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Remove class(es) from the list of classes that
        have generic profile command processing enabled.
        """
        result = self.alter(
            options={"delete:base:generic_command_classes": class_names}
        )
        return self._to_steps(result)

    # ============================================================================
    # Generic Profile Checking Classes
    # ============================================================================
    def add_generic_profile_checking_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Add class(es) to the list of classes that have generic profile checking enabled."""
        result = self.alter(
            options={"base:generic_profile_checking_classes": class_names}
        )
        return self._to_steps(result)

    def remove_generic_profile_checking_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Remove class(es) from the list of classes that have generic profile checking enabled."""
        result = self.alter(
            options={"delete:base:generic_profile_checking_classes": class_names}
        )
        return self._to_steps(result)

    # ============================================================================
    # Generic Profile Sharing Classes
    # ============================================================================
    def add_generic_profile_sharing_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Add class(es) to the list of classes that are eligible for
        general resource profile sharing in common storage.
        """
        result = self.alter(
            options={"base:generic_profile_sharing_classes": class_names}
        )
        return self._to_steps(result)

    def remove_generic_profile_sharing_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Remove class(es) from the list of classes that are eligible
        for general resource profile sharing in common storage.
        """
        result = self.alter(
            options={"delete:base:generic_profile_sharing_classes": class_names}
        )
        return self._to_steps(result)

    # ============================================================================
    # Global Access Classes
    # ============================================================================
    def add_global_access_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Add class(es) to the list of classes eligible for global access checking."""
        return self.alter(options={"base:global_access_classes": class_names})

    def remove_global_access_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Remove class(es) from the list of classes eligible for global access checking."""
        result = self.alter(options={"delete:base:global_access_classes": class_names})
        return self._to_steps(result)

    # ============================================================================
    # Raclist Classes
    # ============================================================================
    def add_raclist_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Add class(es) to list of classes that have in-storage profile sharing activated."""
        result = self.alter(options={"base:raclist": class_names})
        return self._to_steps(result)

    def remove_raclist_classes(
        self, class_names: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """
        Remove class(es) from the list of classes that have in-storage profile sharing activated.
        """
        result = self.alter(options={"delete:base:raclist": class_names})
        return self._to_steps(result)

    # ============================================================================
    # Base Functions
    # ============================================================================
    def list_racf_options(self, options_only: bool = False) -> Union[dict, bytes]:
        """List RACF options."""
        self._build_segment_trait_dictionary({"base:list": True})
        setropts_request = SetroptsRequest()
        self._add_traits_directly_to_request_xml_with_no_segments(setropts_request)
        result = self._extract_and_check_result(setropts_request)
        if options_only:
            return self._get_profile(result)
        return result

    def alter(self, options: dict = {}) -> Union[dict, bytes]:
        """Update RACF options."""
        self._build_segment_trait_dictionary(options)
        setropts_request = SetroptsRequest()
        self._add_traits_directly_to_request_xml_with_no_segments(
            setropts_request, alter=True
        )
        return self._make_request(setropts_request)

    # ============================================================================
    # Private/Protected Utility Functions
    # ============================================================================
    def _format_profile(self, result: dict) -> None:
        """Format profile."""
        messages = result["securityResult"]["systemSettings"]["commands"][0]["messages"]
        # Remove leading/trailing whitespace from messages.
        messages = [message.strip() for message in messages]
        # Ensure that the key-value relationship token is consistent.
        # Change key/field names for easy parsing and categorization.
        messages_normalized = (
            "\n".join(messages)
            .replace("CURRENT OPTIONS:\n", "CURRENT OPTIONS: [OPTION] ")
            .replace("ERASE-ON-SCRATCH BY SECURITY LEVEL", "BY SECURITY LEVEL")
            .replace('"NEVER EXPIRES"', "NONE")
            .replace('"CATDSNS WARNING"', "WARNING")
            .replace('"MLS WARNING"', "WARNING")
            .replace('"MLACTIVE FAIL"', "FAIL")
            .replace('"', "")
            .replace("CURRENT OPTIONS:", "CURRENT OPTIONS")
            .replace(" / ", "_")
            .replace(": ", "= ")
            .replace(":\n", "=\n")
            .replace(" -- ", "--")
            .replace("LOGOPTIONS", "LOG")
            .replace("SUCCESSES CLASSES", "SUCCESS CLASSES")
            .replace("FAILURES CLASSES", "FAILURE CLASSES")
            .replace("SECLEVELAUDIT IS INACTIVE", "SECLEVELAUDIT IS NONE")
            .replace(
                "DATA SET MODELLING IS BEING DONE FOR GDGS.",
                "GENERATION DATA GROUP DATA SET MODELLING = TRUE",
            )
            .replace(
                "DATA SET MODELLING NOT BEING DONE FOR GDGS.",
                "GENERATION DATA GROUP DATA SET MODELLING = FALSE",
            )
            .replace(
                "SINGLE LEVEL NAMES NOT ALLOWED",
                "DATA SET SINGLE LEVEL NAME PREFIX = NONE",
            )
            .replace(" ARE NOT ALLOWED", " = FALSE")
            .replace(" ARE ALLOWED", " = TRUE")
            .replace(" NOT ALLOWED", " = FALSE")
            .replace(" IS ACTIVE", " = TRUE")
            .replace(" IS INACTIVE", " = FALSE")
            .replace(" IN EFFECT IS ", " = ")
            .replace(" IS IN EFFECT FOR THE", " = ")
            .replace(" IS IN EFFECT", " = TRUE")
            .replace(" IS NOT IN EFFECT", " = FALSE")
            .replace(" IS BEING DONE FOR", " = ")
            .replace(" IS BEING DONE", " = TRUE")
            .replace(" NOT BEING DONE FOR ", "")
            .replace(" IS ", " = ")
            .replace("TRUE,", "TRUE.")
            .replace("FUNCTION.", "")
            .replace(" OPTION =", " =")
            .replace("AUTOMATIC DATASET PROTECTION", "AUTOMATIC DATA SET PROTECTION")
            .replace("ENHANCED GENERIC NAMING", "DATA SET ENHANCED GENERIC NAMING")
            .replace("REAL DATA SET NAMES", "REAL DATA SET NAMES LOGGING")
            .replace("PROTECT-ALL", "DATA SET PROTECT-ALL")
            .replace("SECURITY RETENTION PERIOD", "DATA SET SECURITY RETENTION PERIOD")
            .replace("ERASE-ON-SCRATCH", "DATA SET ERASE-ON-SCRATCH")
            .replace("SINGLE LEVEL NAME PREFIX", "DATA SET SINGLE LEVEL NAME PREFIX")
            .replace("CATALOGUED DATA SETS ONLY", "CATALOGUED DATA SET ACCESS ONLY")
            .replace("ADDCREATOR", "DATA SET ADD CREATOR")
            .replace(
                "INACTIVE USERIDS ARE NOT BEING AUTOMATICALLY REVOKED.",
                "REVOKE USERIDS REVOKE AFTER = N/A",
            )
            .replace(
                "INACTIVE USERIDS ARE BEING AUTOMATICALLY REVOKED AFTER ",
                "REVOKE USERIDS REVOKE AFTER = ",
            )
            .replace("= =", "=")
            .replace("ATTRIBUTES = INITSTATS", "ATTRIBUTES = INITIALIZATION-STATISTICS")
            .replace("(PROGRAM", "(PROGRAM-CONTROL")
            .replace("TERMINAL(", "TERMINAL-ACCESS(")
            .replace("PASSWORD PROCESSING OPTIONS=", "")
            .replace(
                "THE ACTIVE PASSWORD ENCRYPTION ALGORITHM",
                "ACTIVE PASSWORD ENCRYPTION ALGORITHM",
            )
            .replace("SPECIAL CHARACTERS", "SPECIAL CHARACTERS ALLOWED")
            .replace(
                "NO PASSWORD HISTORY BEING MAINTAINED.",
                "PASSWORD HISTORY GENERATIONS = 0",
            )
            .replace(
                "GENERATIONS OF PREVIOUS PASSWORDS BEING MAINTAINED.",
                "PASSWORD HISTORY GENERATIONS = [TOKEN 0]",
            )
            .replace(
                "USERIDS NOT BEING AUTOMATICALLY REVOKED",
                "MAX UNSUCCESSFUL PASSWORD ATTEMPTS = NONE",
            )
            .replace(
                "CONSECUTIVE UNSUCCESSFUL PASSWORD ATTEMPTS, A USERID WILL BE REVOKED.",
                "MAX UNSUCCESSFUL PASSWORD ATTEMPTS = [TOKEN 1]",
            )
            .replace(
                "CONSECUTIVE UNSUCCESSFUL PASSWORD ATTEMPTS,\nA USERID WILL BE REVOKED.",
                "MAX UNSUCCESSFUL PASSWORD ATTEMPTS = [TOKEN 1]",
            )
            .replace(
                "NO INSTALLATION PASSWORD SYNTAX RULES ARE PRESENT.",
                "SYNTAX RULES = NONE",
            )
            .replace("INSTALLATION PASSWORD SYNTAX RULES", "SYNTAX RULES")
            .replace("W-NOVOWEL", "W-NO VOWEL")
            .replace("L-ALPHANUM", "L-ALPHANUMERIC")
            .replace("GLOBAL=YES RACLIST ONLY", "GLOBAL RACLIST ONLY CLASSES")
            .replace("GENLIST CLASSES =", "GENERIC PROFILE SHARING CLASSES =")
            .replace("SETR RACLIST CLASSES =", "RACLIST CLASSES =")
            .replace("KERBLVL =", "KERBEROS ENCRYPTION LEVEL =")
            .replace("JES-BATCHALLRACF", "JES-BATCH ALL")
            .replace("JES-XBMALLRACF", "JES-EXECUTION BATCH MONITOR ALL")
            .replace("JES-EARLYVERIFY", "JES-EARLY VERIFY")
            .replace("USER-ID FOR JES UNDEFINEDUSER", "JES UNDEFINED USER")
            .replace("USER-ID FOR JES NJEUSERID", "JES NETWORK USER")
            .replace("NO WRITE-DOWN", "MULTI-LEVEL NO WRITE-DOWN")
            .replace("SECLEVELAUDIT", "SECURITY LABEL LEVEL AUDITING")
            .replace("SECURITY LEVEL FOR AUDITING", "SECURITY LABEL LEVEL AUDITING")
            .replace("SECLABEL AUDIT", "SECURITY LABEL LABEL AUDITING")
            .replace("SECLABEL CONTROL", "SECURITY LABEL ALLOW CONTROL FOR READ ACCESS")
            .replace("COMPATIBILITY MODE", "SECURITY LABEL COMPATIBILITY MODE")
            .replace("GENERIC OWNER ONLY", "GENERIC RULES RESTRICT GENERIC OWNER")
            .replace(
                "LIST OF GROUPS ACCESS CHECKING",
                "GROUP RULES LIST OF GROUPS ACCESS CHECKING",
            )
            .replace(
                "PARTNER LU-VERIFICATION SESSIONKEY INTERVAL MAXIMUM/DEFAULT",
                "VTAM SESSION KEY VERIFICATION INTERVAL",
            )
            .replace(
                "PARTNER LU-VERIFICATION SESSIONKEY INTERVAL DEFAULT",
                "VTAM SESSION KEY VERIFICATION INTERVAL",
            )
            .replace("APPLAUDIT", "VTAM APPC TRANSACTION AUDIT")
            .splitlines()
        )
        # Merge multi-line fields into single line based on key-value relationship token.
        messages_with_merged_lists = []
        for line in messages_normalized:
            if "=" not in line and len(messages_with_merged_lists) != 0:
                messages_with_merged_lists[-1] += " " + line
            else:
                messages_with_merged_lists.append(line)
        # Build initial mapping according to key-value pair relationships
        profile_raw = []
        for line in messages_with_merged_lists:
            tokens = line.split("=")
            key_raw = tokens[0].strip()
            value_raw = "=".join(tokens[1:]).strip().rstrip(".")
            profile_raw.append((key_raw, value_raw))
        class_list_fields = [
            "STATISTICS",
            "ACTIVE CLASSES",
            "GENERIC PROFILE CLASSES",
            "GENERIC COMMAND CLASSES",
            "GENERIC PROFILE SHARING CLASSES",
            "GLOBAL CHECKING CLASSES",
            "RACLIST CLASSES",
            "GLOBAL RACLIST ONLY CLASSES",
            "LOG ALWAYS CLASSES",
            "LOG NEVER CLASSES",
            "LOG SUCCESS CLASSES",
            "LOG FAILURE CLASSES",
            "LOG DEFAULT CLASSES",
            "AUDIT CLASSES",
        ]
        generic_list_fields = ["DEFAULT RVARY PASSWORD"]
        password_processing_options = [
            "ACTIVE PASSWORD ENCRYPTION ALGORITHM",
            "PASSWORD CHANGE INTERVAL",
            "PASSWORD MINIMUM CHANGE INTERVAL",
            "MIXED CASE PASSWORD SUPPORT",
            "SPECIAL CHARACTERS ALLOWED",
            "PASSWORD HISTORY GENERATIONS",
            "MAX UNSUCCESSFUL PASSWORD ATTEMPTS",
            "PASSWORD EXPIRATION WARNING LEVEL",
            "SYNTAX RULES",
            "LEGEND",
        ]
        generic_subfield_map = {
            "JES": "jes",
            "REVOKE USERIDS": "revokeUserids",
            "MULTI-LEVEL": "multiLevelSecurity",
            "LANGUAGE DEFAULT": "languageDefaults",
            "DATA SET": "dataSets",
            "SECURITY LABEL": "securityLabels",
            "VTAM": "vtam",
            "KERBEROS": "kerberos",
            "GENERIC RULES": "genericRules",
            "GROUP RULES": "groupRules",
        }
        generic_subsubfield_map = {
            "MULTI-LEVEL": [
                "NO WRITE-DOWN",
                "SECURE",
                "ACTIVE",
            ],
            "DATA SET": ["CATALOGUED DATA SET ACCESS ONLY", "ERASE-ON-SCRATCH"],
        }
        profile = {}
        for key_raw, value_raw in profile_raw:
            if "[TOKEN " in value_raw:
                (key_raw, value_raw) = self.__fix_key_value_raw(key_raw, value_raw)
            key = self._profile_field_to_camel_case("base", key_raw.lower())
            if key_raw == "ATTRIBUTES":
                self.__add_attributes(profile, key, value_raw)
                continue
            elif key_raw in class_list_fields:
                self.__add_classes_list(profile, key, value_raw)
                continue
            elif key_raw in generic_list_fields:
                value = self.__to_list(value_raw)
                if key in profile:
                    profile[key] += value
                else:
                    profile[key] = value
                continue
            elif key_raw in password_processing_options:
                self.__add_password_processing_options(profile, key, value_raw)
                continue
            elif self.__is_generic_subfield(key_raw, generic_subfield_map):
                self.__add_generic_subfield(
                    profile,
                    generic_subfield_map,
                    generic_subsubfield_map,
                    key_raw,
                    value_raw,
                )
                continue
            value = self._cast_from_str(value_raw)
            profile[key] = value
        del result["securityResult"]["systemSettings"]["commands"][0]["messages"]
        result["securityResult"]["systemSettings"]["commands"][0]["profiles"] = [
            profile
        ]

    def __fix_key_value_raw(self, key_raw: str, value_raw: str) -> Tuple[str, str]:
        """Normalize raw key-value pair."""
        key_tokens = key_raw.split()
        value_index = int(value_raw.split()[-1].rstrip("]"))
        value_raw = key_tokens[value_index]
        key_raw = " ".join(key_tokens[value_index + 1 :])
        return (key_raw, value_raw)

    def __add_attributes(self, profile: dict, key: str, value_raw: str):
        """Add attributes to the profile field"""
        if key not in profile:
            profile[key] = {}
        tokens = value_raw.split()
        for token in tokens:
            if "(" in token:
                subtokens = token.split("(")
                if subtokens[0] == "WHEN":
                    subsubtokens = subtokens[1].split("--")
                    attribute = subsubtokens[0]
                    value = subsubtokens[-1].rstrip(")").lower()
                elif subtokens[0] == "NOWHEN":
                    attribute = subtokens[1].rstrip(")")
                    value = False
                else:
                    attribute = subtokens[0]
                    value = subtokens[1].rstrip(")").lower()
            elif token[:2] == "NO":
                attribute = token[2:]
                value = False
            else:
                attribute = token
                value = True
            match attribute:
                case "SAUDIT":
                    attribute = "SPECIAL AUDIT"
                case "CMDVIOL":
                    attribute = "LOG COMMAND VIOLATIONS"
                case "OPERAUDIT":
                    attribute = "OPERATIONS AUDIT"
            attribute = self._profile_field_to_camel_case("base", attribute.lower())
            profile[key][attribute] = value

    def __to_list(self, value_raw: str, n: int = 1) -> List[str]:
        """Convert space delimited list into a list."""
        if value_raw == "NONE":
            return []
        tokens = value_raw.split()
        return [
            self._cast_from_str(" ".join(tokens[i : i + n]))
            for i in range(0, len(tokens), n)
        ]

    def __add_classes_list(self, profile: dict, class_key: str, value_raw: str) -> None:
        """Add a class list to profile."""
        if "classes" not in profile:
            profile["classes"] = {}
        class_key = class_key.replace("Classes", "")
        profile["classes"][class_key] = self.__to_list(value_raw)

    def __add_password_processing_options(
        self, profile: dict, key: str, data: str
    ) -> None:
        """Add password processing options to profile."""
        if "passwordProcessingOptions" not in profile:
            profile["passwordProcessingOptions"] = {}
        if key == "syntaxRules":
            try:
                profile["passwordProcessingOptions"][key]["rules"]
            except KeyError:
                profile["passwordProcessingOptions"][key] = {"rules": []}
            if data == "NONE":
                return
            rule_tokens = [
                rule_token for rule_token in data.split("RULE") if rule_token != ""
            ]
            for rule_token in rule_tokens:
                length_chars = rule_token.lower().split("length(")[1].split(")")[0]
                if ":" in length_chars:
                    min_length = self._cast_from_str(length_chars.split(":")[0])
                    max_length = self._cast_from_str(length_chars.split(":")[1])
                else:
                    min_length = self._cast_from_str(length_chars)
                    max_length = self._cast_from_str(length_chars)
                chars = rule_token[-1 * max_length :].strip()
                profile["passwordProcessingOptions"][key]["rules"].append(
                    {
                        "minLength": min_length,
                        "maxLength": max_length,
                        "content": chars,
                    }
                )
            return
        elif key == "legend":
            legend = {}
            legend_tokens = data.split("-")
            for i in range(0, len(legend_tokens) - 1):
                legend_key = legend_tokens[i].split()[-1]
                if i + 1 == len(legend_tokens) - 1:
                    legend_value = legend_tokens[-1]
                else:
                    legend_value = " ".join(legend_tokens[i + 1].split()[:-1])
                legend[legend_key] = legend_value.lower()
            profile["passwordProcessingOptions"]["syntaxRules"][key] = legend
            return
        profile["passwordProcessingOptions"][key] = self._cast_from_str(data)

    def __is_generic_subfield(self, key_raw: str, generic_subfield_map: dict):
        """Check if a subfield is considered a generic subfield."""
        for subfield in generic_subfield_map.keys():
            if subfield in key_raw:
                return True
        return False

    def __add_generic_subfield(
        self,
        profile: dict,
        generic_subfield_map: dict,
        generic_subsubfield_map: dict,
        key_raw: str,
        value_raw: str,
    ) -> None:
        """Add a generic subfield to profile."""
        subdictionary = {}
        for subfield_token in generic_subsubfield_map.keys():
            if subfield_token in key_raw:
                for subsubfield_token in generic_subsubfield_map[subfield_token]:
                    if subsubfield_token in key_raw:
                        if "TRUE. CURRENT OPTIONS" in value_raw:
                            value_tokens = value_raw.split(". CURRENT OPTIONS")
                            subdictionary["enabled"] = self._cast_from_str(
                                value_tokens[0]
                            )
                            subdictionary[
                                "options"
                            ] = self.__process_generic_subsubfield_options(
                                value_tokens[1]
                            )
                        else:
                            subdictionary["enabled"] = self._cast_from_str(value_raw)
                            subdictionary["options"] = {}
        for subfield_token in generic_subfield_map.keys():
            if subfield_token in key_raw:
                key = generic_subfield_map[subfield_token]
                subkey_raw = key_raw.replace(subfield_token, "")
        if key not in profile:
            profile[key] = {}
        subkey = self._profile_field_to_camel_case("base", subkey_raw.lower())
        if subdictionary:
            profile[key][subkey] = subdictionary
        else:
            profile[key][subkey] = self._cast_from_str(value_raw)

    def __process_generic_subsubfield_options(self, subsubfield_options: str) -> dict:
        # Build a dictionary that represents the options associated with a generic subfield.
        subsubfield_option_tokens = subsubfield_options.split("[OPTION]")[1:]
        options_dictionary = {}
        for option in subsubfield_option_tokens:
            option_tokens = option.split("=")
            key = self._profile_field_to_camel_case(
                "base", option_tokens[0].strip().lower()
            )
            value = self._cast_from_str(option_tokens[1].strip())
            options_dictionary[key] = value
        return options_dictionary



"""Exception to use when data returned by IRRSMO00 indicates that the request failed."""


class SecurityRequestError(Exception):
    """
    Raised when the return code of a security result returned by IRRSMO00 is NOT equal to 0.
    """

    def __init__(self, result: dict) -> None:
        self.message = "Security request made to IRRSMO00 failed."
        self.result = result
        self.message += (
            "\n\nSee results dictionary "
            + f"'{self.__class__.__name__}.result' for more details."
        )
        self.message = f"({self.__class__.__name__}) {self.message}"

    def __str__(self) -> str:
        return self.message

    def contains_error_message(
        self, security_definition_tag: str, error_message_id: str
    ):
        commands = self.result["securityResult"][security_definition_tag].get(
            "commands"
        )
        if not isinstance(commands, list):
            return False
        messages = commands[0].get("messages", [])
        if error_message_id in "".join(messages):
            return True
        else:
            return False



"""Interface to irrsmo00.dll."""
import platform

try:
    from cpyracf import call_irrsmo00
except ImportError as import_error:
    if platform.system() == "OS/390":
        raise import_error

    # Ignore import of extension on non-z/OS platforms to allow for unit testing off platform.
    def call_irrsmo00() -> None:
        return None


class IRRSMO00:
    """Interface to irrsmo00 callable service through cpyracf Python extension."""

    def __init__(self) -> None:
        # Initialize size of output buffer
        self.buffer_size = 100000

    def call_racf(self, request_xml: bytes, precheck: bool = False) -> str:
        """Make request to call_irrsmo00 in the cpyracf Python extension."""
        options = 15 if precheck else 13
        return call_irrsmo00(
            xml_str=request_xml, xml_len=len(request_xml), opts=options
        ).decode("cp1047")



"""Group Administration Request Builder."""

from pyracf.common.security_request import SecurityRequest


class GroupRequest(SecurityRequest):
    """Group Administration Request Builder."""

    def __init__(self, groupid: str, operation: str) -> None:
        super().__init__()
        self._security_definition.tag = "group"
        self._security_definition.attrib = {
            "name": groupid,
            "operation": operation,
            "requestid": "GroupRequest",
        }



"""Group Administration."""

from typing import List, Union

from pyracf.common.add_operation_error import AddOperationError
from pyracf.common.alter_operation_error import AlterOperationError
from pyracf.common.security_admin import SecurityAdmin
from pyracf.common.security_request_error import SecurityRequestError

from .group_request import GroupRequest


class GroupAdmin(SecurityAdmin):
    """Group Administration."""

    def __init__(
        self,
        debug: bool = False,
        generate_requests_only: bool = False,
        update_existing_segment_traits: Union[dict, None] = None,
        replace_existing_segment_traits: Union[dict, None] = None,
        additional_secret_traits: Union[List[str], None] = None,
    ) -> None:
        self._valid_segment_traits = {
            "base": {
                "base:installation_data": "racf:data",
                "base:data_set_model": "racf:model",
                "base:owner": "racf:owner",
                "base:superior_group": "racf:supgroup",
                "base:terminal_universal_access": "racf:termuacc",
                "base:universal": "racf:universl",
            },
            "dfp": {
                "dfp:data_application": "dataappl",
                "dfp:data_class": "dataclas",
                "dfp:management_class": "mgmtclas",
                "dfp:storage_class": "storclas",
            },
            "omvs": {
                "omvs:auto_gid": "racf:autogid",
                "omvs:gid": "gid",
                "omvs:shared": "racf:shared",
            },
            "ovm": {"ovm:gid": "racf:gid"},
            "tme": {"tme:roles": "racf:roles"},
        }
        super().__init__(
            "group",
            debug=debug,
            generate_requests_only=generate_requests_only,
            update_existing_segment_traits=update_existing_segment_traits,
            replace_existing_segment_traits=replace_existing_segment_traits,
            additional_secret_traits=additional_secret_traits,
        )

    # ============================================================================
    # Group Special
    # ============================================================================
    def has_group_special_authority(
        self, group: str, userid: str
    ) -> Union[bool, bytes]:
        """Check if a user is connected to a group with group special authority."""
        profile = self.extract(group, profile_only=True)
        return self.__has_connect_attribute(userid, profile, "special")

    # ============================================================================
    # Group Operations
    # ============================================================================
    def has_group_operations_authority(
        self, group: str, userid: str
    ) -> Union[bool, bytes]:
        """Check if a user is connected to a group with group operations authority."""
        profile = self.extract(group, profile_only=True)
        return self.__has_connect_attribute(userid, profile, "operations")

    # ============================================================================
    # Group Auditor
    # ============================================================================
    def has_group_auditor_authority(
        self, group: str, userid: str
    ) -> Union[bool, bytes]:
        """Check if a user is connected to a group with group auditor authority."""
        profile = self.extract(group, profile_only=True)
        return self.__has_connect_attribute(userid, profile, "auditor")

    # ============================================================================
    # Group Access
    # ============================================================================
    def has_group_access_attribute(self, group: str, userid: str) -> Union[bool, bytes]:
        """Check if a user is connected to a group with the group access attribute."""
        profile = self.extract(group, profile_only=True)
        return self.__has_connect_attribute(userid, profile, "grpacc")

    # ============================================================================
    # OMVS GID
    # ============================================================================
    def get_omvs_gid(self, group: str) -> Union[int, bytes]:
        """Get a group's OMVS GID."""
        profile = self.extract(group, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "gid")

    def set_omvs_gid(self, group: str, gid: int) -> Union[dict, bytes]:
        """Set a group's OMVS GID."""
        result = self.alter(group, traits={"omvs:gid": gid})
        return self._to_steps(result)

    # ============================================================================
    # OVM GID
    # ============================================================================
    def get_ovm_gid(self, group: str) -> Union[int, bytes]:
        """Get a group's OVM GID."""
        profile = self.extract(group, segments=["ovm"], profile_only=True)
        return self._get_field(profile, "ovm", "gid")

    def set_ovm_gid(self, group: str, gid: int) -> Union[dict, bytes]:
        """Set a group's OVM GID."""
        result = self.alter(group, traits={"ovm:gid": gid})
        return self._to_steps(result)

    # ============================================================================
    # Base Functions
    # ============================================================================
    def add(self, group: str, traits: dict = {}) -> Union[dict, bytes]:
        """Create a new group."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            group_request = GroupRequest(group, "set")
            self._build_xml_segments(group_request)
            return self._make_request(group_request)
        try:
            self.extract(group)
        except SecurityRequestError as exception:
            if not exception.contains_error_message(self._profile_type, "ICH51003I"):
                raise exception
            self._build_segment_trait_dictionary(traits)
            group_request = GroupRequest(group, "set")
            self._build_xml_segments(group_request)
            return self._make_request(group_request)
        raise AddOperationError(group, self._profile_type)

    def alter(self, group: str, traits: dict) -> Union[dict, bytes]:
        """Alter an existing group."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            group_request = GroupRequest(group, "set")
            self._build_xml_segments(group_request, alter=True)
            return self._make_request(group_request, irrsmo00_precheck=True)
        try:
            self.extract(group)
        except SecurityRequestError:
            raise AlterOperationError(group, self._profile_type)
        self._build_segment_trait_dictionary(traits)
        group_request = GroupRequest(group, "set")
        self._build_xml_segments(group_request, alter=True)
        return self._make_request(group_request, irrsmo00_precheck=True)

    def extract(
        self, group: str, segments: List[str] = [], profile_only: bool = False
    ) -> Union[dict, bytes]:
        """Extract a group's profile."""
        self._build_segment_dictionary(segments)
        group_request = GroupRequest(group, "listdata")
        self._build_xml_segments(group_request, extract=True)
        result = self._extract_and_check_result(group_request)
        if profile_only:
            return self._get_profile(result)
        return result

    def delete(self, group: str) -> Union[dict, bytes]:
        """Delete a group."""
        group_request = GroupRequest(group, "del")
        return self._make_request(group_request)

    # ============================================================================
    # Private/Protected Utility Functions
    # ============================================================================
    def _format_profile(self, result: dict) -> None:
        """Format profile extract data into a dictionary."""
        messages = result["securityResult"]["group"]["commands"][0]["messages"]
        indexes = [
            i
            for i in range(len(messages) - 1)
            if messages[i] and "INFORMATION FOR GROUP " in messages[i]
        ]
        indexes.append(len(messages))
        profiles = []
        for i in range(len(indexes) - 1):
            profile = self._format_profile_generic(
                messages[indexes[i] : indexes[i + 1]]
            )
            if "subgroup(s)" in profile.keys():
                profile["subgroup(s)"] = profile["subgroups"]
            profiles.append(profile)

        # Post processing
        del result["securityResult"]["group"]["commands"][0]["messages"]
        result["securityResult"]["group"]["commands"][0]["profiles"] = profiles

    def __has_connect_attribute(
        self, userid: str, profile: dict, attribute: str
    ) -> bool:
        """check if a user has a connect attribute in a group profile."""
        if profile["base"]["users"] is None:
            return False
        connect_profile = [
            user
            for user in profile["base"]["users"]
            if user["userid"].lower() == userid.lower()
        ]
        if connect_profile:
            if attribute in connect_profile[0]["connectAttributes"]:
                return True
        return False



"""Connection Request Builder."""

from pyracf.common.security_request import SecurityRequest


class ConnectionRequest(SecurityRequest):
    """Connection Request Builder."""

    def __init__(
        self,
        userid: str,
        groupname: str,
        operation: str,
    ) -> None:
        super().__init__()
        self._security_definition.tag = "groupconnection"
        self._security_definition.attrib.update(
            {
                "name": userid,
                "group": groupname,
                "operation": operation,
                "requestid": "ConnectionRequest",
            }
        )



"""Generic Security Result Parser."""

import re
from xml.etree.ElementTree import Element  # Only used for type hints.

import defusedxml.ElementTree as XMLParser


class SecurityResult:
    """Generic Security Result Parser."""

    def __init__(self, result_xml: str) -> None:
        self.__result = XMLParser.fromstring(result_xml)
        self.__result_dictionary = {"securityResult": {}}
        self.__extract_results()

    def __extract_results(self) -> None:
        """Extract XML results into a dictionary."""
        self.__definition = self.__result[0]
        return_code = int(self.__result[1].text)
        reason_code = int(self.__result[2].text)
        self.__definition.attrib["requestId"] = self.__definition.attrib["requestid"]
        del self.__definition.attrib["requestid"]
        definition_tag = self.__to_pascal_case(self.__definition.tag.split("}")[-1])
        self.__result_dictionary["securityResult"][
            definition_tag
        ] = self.__definition.attrib
        self.__definition_dictionary = self.__result_dictionary["securityResult"][
            definition_tag
        ]
        filter_out_extra_messages = False
        if return_code == 0 and self.__definition_dictionary["operation"] == "listdata":
            filter_out_extra_messages = True
        try:
            if self.__definition[0].tag.split("}")[-1] == "info":
                self.__extract_info()
            if self.__definition[0].tag.split("}")[-1] == "error":
                self.__extract_error()
            else:
                self.__extract_commands(filter_out_extra_messages)
        except IndexError:
            # Index Error indicates that there is no
            # additional information to extract from the definition.
            pass
        self.__result_dictionary["securityResult"]["returnCode"] = return_code
        self.__result_dictionary["securityResult"]["reasonCode"] = reason_code

    def __extract_info(self) -> None:
        """Extract info section from XML into a list."""
        self.__definition_dictionary["info"] = []
        info = self.__definition_dictionary["info"]
        while self.__definition[0].tag.split("}")[-1] == "info":
            item = self.__definition[0]
            if item.tag.split("}")[-1] != "info":
                return
            info.append(item.text)
            self.__definition.remove(item)

    def __extract_commands(self, filter_out_extra_messages: bool) -> None:
        """Extract commands section from XML into a list."""
        self.__definition_dictionary["commands"] = []
        commands = self.__definition_dictionary["commands"]
        for command in self.__definition:
            self.__extract_command(commands, command, filter_out_extra_messages)

    def __extract_command(
        self,
        commands: dict,
        command: Element,
        filter_out_extra_messages: bool,
    ) -> None:
        command_dictionary = {}
        commands.append(command_dictionary)
        message_id_regex = r"[A-Z]{3}[0-9]{5}[A-Z]"
        for item in command:
            item_tag = self.__to_pascal_case(item.tag.split("}")[-1])
            if item_tag == "message":
                if "messages" not in command_dictionary:
                    command_dictionary["messages"] = []
                if item.text:
                    if filter_out_extra_messages and re.match(
                        message_id_regex, item.text
                    ):
                        continue
                command_dictionary["messages"].append(item.text)
            else:
                try:
                    command_dictionary[item_tag] = int(item.text)
                except ValueError:
                    command_dictionary[item_tag] = item.text

    def __extract_error(self) -> None:
        """Extract error section from XML into a dictionary."""
        self.__definition_dictionary["error"] = {}
        error = self.__definition[0]
        for item in error:
            item_tag = self.__to_pascal_case(item.tag.split("}")[-1])
            try:
                self.__definition_dictionary["error"][item_tag] = int(item.text)
            except ValueError:
                self.__definition_dictionary["error"][item_tag] = item.text

    def __to_pascal_case(self, key: str) -> str:
        """Convert result dictionary keys to pascal case."""
        match (key):
            case "returncode":
                return "returnCode"
            case "reasoncode":
                return "reasonCode"
            case "safreturncode":
                return "safReturnCode"
            case "errorfunction":
                return "errorFunction"
            case "errorcode":
                return "errorCode"
            case "errorreason":
                return "errorReason"
            case "errormessage":
                return "errorMessage"
            case "erroroffset":
                return "errorOffset"
            case "textinerror":
                return "textInError"
            case "groupconnection":
                return "groupConnection"
            case "dataset":
                return "dataSet"
            case "systemsettings":
                return "systemSettings"
            case _:
                return key

    def get_result_dictionary(self) -> dict:
        """Return result dictionary."""
        return self.__result_dictionary



"""Build IRRSMO00 (cpyracf) Python extesion."""

import os

from setuptools import Extension
from setuptools.command import build_ext


def build(setup_kwargs: dict):
    """Python extension build entrypoint."""
    os.environ["_CC_CCMODE"] = "1"
    os.environ["_CXX_CCMODE"] = "1"
    os.environ["_C89_CCMODE"] = "1"
    os.environ["_CC_EXTRA_ARGS"] = "1"
    os.environ["_CXX_EXTRA_ARGS"] = "1"
    os.environ["_C89_EXTRA_ARGS"] = "1"
    os.environ["CC"] = "xlc"
    os.environ["CXX"] = "xlc++"
    setup_kwargs.update(
        {
            "ext_modules": [
                Extension(
                    "cpyracf",
                    sources=["pyracf/common/irrsmo00.c"],
                    extra_compile_args=[
                        "-D_XOPEN_SOURCE_EXTENDED",
                        "-Wc,lp64,langlvl(EXTC99),STACKPROTECT(ALL),",
                        "-qcpluscmt",
                    ],
                )
            ],
            "cmdclass": {"built_ext": build_ext},
        }
    )



"""General Resource Profile Administration."""

from typing import List, Union

from pyracf.common.add_operation_error import AddOperationError
from pyracf.common.alter_operation_error import AlterOperationError
from pyracf.common.security_admin import SecurityAdmin
from pyracf.common.security_request_error import SecurityRequestError

from .resource_request import ResourceRequest


class ResourceAdmin(SecurityAdmin):
    """General Resaurce Profile Administration."""

    def __init__(
        self,
        debug: bool = False,
        generate_requests_only: bool = False,
        update_existing_segment_traits: Union[dict, None] = None,
        replace_existing_segment_traits: Union[dict, None] = None,
        additional_secret_traits: Union[List[str], None] = None,
    ) -> None:
        self._valid_segment_traits = {
            "base": {
                "base:application_data": "racf:appldata",
                "base:audit_alter:": "racf:audaltr",
                "base:audit_control": "racf:audcntl",
                "base:audit_none": "racf:audnone",
                "base:audit_read": "racf:audread",
                "base:audit_update": "racf:audupdt",
                "base:security_categories": "racf:category",
                "base:installation_data": "racf:data",
                "base:model_profile_class": "racf:fclass",
                "base:model_profile_generic": "racf:fgeneric",
                "base:model_profile": "racf:fprofile",
                "base:model_profile_volume": "racf:fvolume",
                "base:global_audit_alter": "racf:gaudaltr",
                "base:global_audit_control": "racf:gaudcntl",
                "base:global_audit_none": "racf:gaudnone",
                "base:global_audit_read": "racf:gaudread",
                "base:global_audit_update": "racf:gaudupdt",
                "base:level": "racf:level",
                "base:member": "racf:member",
                "base:notify_userid": "racf:notify",
                "base:owner": "racf:owner",
                "base:security_label": "racf:seclabel",
                "base:security_level": "racf:seclevel",
                "base:single_data_set_tape_volume": "racf:singldsn",
                "base:time_zone": "racf:timezone",
                "base:tape_vtoc": "racf:tvtoc",
                "base:universal_access": "racf:uacc",
                "base:volumes": "racf:volume",
                "base:warn_on_insufficient_access": "racf:warning",
                "base:terminal_access_allowed_days": "racf:whendays",
                "base:terminal_access_allowed_time": "racf:whentime",
            },
            "cdtinfo": {
                "cdtinfo:case_allowed": "case",
                "cdtinfo:default_racroute_return_code": "defaultrc",
                "cdtinfo:valid_first_characters": "first",
                "cdtinfo:generic_profile_checking": "racf:generic",
                "cdtinfo:generic_profile_sharing": "racf:genlist",
                "cdtinfo:grouping_class_name": "grouping",
                "cdtinfo:key_qualifiers": "keyqual",
                "cdtinfo:manditory_access_control_processing": "macprocessing",
                "cdtinfo:max_length": "maxlenx",
                "cdtinfo:max_length_entityx": "maxlength",
                "cdtinfo:member_class_name": "member",
                "cdtinfo:operations": "operations",
                "cdtinfo:valid_other_characters": "other",
                "cdtinfo:posit_number": "posit",
                "cdtinfo:profiles_allowed": "profilesallowed",
                "cdtinfo:raclist_allowed": "raclist",
                "cdtinfo:send_enf_signal_on_profile_creation": "signal",
                "cdtinfo:security_label_required": "seclabelrequired",
                "cdtinfo:default_universal_access": "defaultuacc",
            },
            "cfdef": {
                "cfdef:custom_field_data_type": "type",
                "cfdef:valid_first_characters": "first",
                "cfdef:help_text": "help",
                "cfdef:list_heading_text": "listhead",
                "cfdef:mixed_case_allowed": "mixed",
                "cfdef:min_numeric_value": "minvalue",
                "cfdef:max_field_length": "mxlength",
                "cfdef:max_numeric_value": "maxvalue",
                "cfdef:valid_other_characters": "other",
                "cfdef:validation_rexx_exec": "racf:cfvalrx",
            },
            "dlfdata": {
                "dlfdata:job_names": "racf:jobname",
                "dlfdata:retain": "racf:retain",
            },
            "eim": {
                "eim:domain_distinguished_name": "domaindn",
                "eim:kerberos_registry": "kerberg",
                "eim:local_registry": "localreg",
                "eim:options": "options",
                "eim:x509_registry": "X509reg",
            },
            "kerb": {
                "kerb:validate_addresses": "checkaddrs",
                "kerb:default_ticket_life": "deftktlife",
                "kerb:encryption_algorithms": "encrypt",
                "kerb:realm_name": "kerbname",
                "kerb:max_ticket_life": "maxtktlf",
                "kerb:min_ticket_life": "mintklife",
                "kerb:password": "password",
            },
            "icsf": {
                "icsf:symmetric_export_certificates": "symexportcert",
                "icsf:exportable_public_keys": "symexportable",
                "icsf:symmetric_export_public_keys": "symexportkey",
                "icsf:symmetric_cpacf_rewrap": "symcpacfwrap",
                "icsf:symmetric_cpacf_rewrap_return": "symcpacfret",
                "icsf:asymetric_key_usage": "asymusage",
            },
            "ictx": {
                "ictx:use_identity_map": "domap",
                "ictx:require_identity_mapping": "mapreq",
                "ictx:identity_map_timeout": "maptimeo",
                "ictx:cache_application_provided_identity_map": "usemap",
            },
            "idtparms": {
                "idtparms:token": "sigtoken",
                "idtparms:sequence_number": "sigseqnum",
                "idtparms:category": "sigcat",
                "idtparms:signature_algorithm": "sigalg",
                "idtparms:identity_token_timeout": "idttimeout",
                "idtparms:use_for_any_application": "anyappl",
            },
            "jes": {"jes:key_label": "racf:keylabel"},
            "mfpolicy": {
                "mfpolicy:factors": "racf:factors",
                "mfpolicy:token_timeout": "racf:timeout",
                "mfpolicy:reuse_token": "racf:reuse",
            },
            "proxy": {
                "proxy:bind_distinguished_name": "binddn",
                "proxy:bind_password": "bindpw",
                "proxy:ldap_host": "ldaphost",
            },
            "session": {
                "session:security_checking_level": "racf:convsec",
                "session:session_key_interval": "racf:interval",
                "session:locked": "racf:lock",
                "session:session_key": "racf:sesskey",
            },
            "sigver": {
                "sigver:fail_program_load_condition": "failload",
                "sigver:log_signature_verification_events": "sigaudit",
                "sigver:signature_required": "sigrequired",
            },
            "ssignon": {
                "ssignon:encrypt_legacy_pass_ticket_key": "racf:keycrypt",
                "ssignon:enhanced_pass_ticket_label": "ptkeylab",
                "ssignon:enhanced_pass_ticket_type": "pttype",
                "ssignon:enhanced_pass_ticket_timeout": "pttimeo",
                "ssignon:enhanced_pass_ticket_replay": "ptreplay",
                "ssignon:legacy_pass_ticket_label": "racf:keylabel",
                "ssignon:mask_legacy_pass_ticket_key": "racf:keymask",
            },
            "stdata": {
                "stdata:group": "racf:group",
                "stdata:privileged": "racf:privlege",
                "stdata:trace": "racf:trace",
                "stdata:trusted": "racf:trusted",
                "stdata:user": "racf:user",
            },
            "svfmr": {
                "svfmr:parameter_list_name": "racf:parmname",
                "svfmr:script_name": "racf:script",
            },
            "tme": {
                "tme:children": "racf:children",
                "tme:groups": "racf:groups",
                "tme:parent": "racf:parent",
                "tme:resource": "racf:resource",
                "tme:roles": "racf:roles",
            },
        }
        self._extracted_key_value_pair_segment_traits_map = {
            "cdtinfo": {
                "case": "caseAllowed",
                "defaultrc": "defaultRacrouteReturnCode",
                "first": "validFirstCharacters",
                "generic": "genericProfileChecking",
                "genlist": "genericProfileSharing",
                "group": "groupingClassName",
                "keyqualifiers": "keyQualifiers",
                "macprocessing": "manditoryAccessControlProcessing",
                "maxlenx": "maxLength",
                "maxlength": "maxLengthEntityx",
                "member": "memberClassName",
                "operations": "operations",
                "other": "validOtherCharacters",
                "posit": "positNumber",
                "profilesallowed": "profilesAllowed",
                "raclist": "raclistAllowed",
                "signal": "sendEnfSignalOnProfileCreation",
                "seclabelsrequired": "securityLabelsRequired",
                "defaultuacc": "defaultUniversalAccess",
            },
            "cfdef": {
                "type": "customFieldDataType",
                "first": "validFirstCharacters",
                "help": "helpText",
                "listhead": "listHeadingText",
                "mixed": "mixedCaseAllowed",
                "minvalue": "minNumericValue",
                "mxlength": "maxFieldLength",
                "maxvalue": "maxNumericValue",
                "other": "validOtherCharacters",
                "cfvalrx": "validationRexxExec",
            },
            "kerb": {
                "checkaddrs": "validateAddresses",
                "deftktlife": "defaultTicketLife",
                "encrypt": "encryptionAlgorithms",
                "kerbname": "realmName",
                "maxtktlf": "maxTicketLife",
                "mintklife": "minTicketLife",
            },
            "session": {
                "convsec": "securityCheckingLevel",
                "interval": "sessionKeyInterval",
                "lock": "locked",
                "sesskey": "sessionKey",
            },
            "sigver": {
                "failload": "failProgramLoadCondition",
                "sigaudit": "logSignatureVerificationEvents",
                "sigrequired": "signatureRequired",
            },
        }
        super().__init__(
            "resource",
            debug=debug,
            generate_requests_only=generate_requests_only,
            update_existing_segment_traits=update_existing_segment_traits,
            replace_existing_segment_traits=replace_existing_segment_traits,
            additional_secret_traits=additional_secret_traits,
        )

    # ============================================================================
    # Universal Access
    # ============================================================================
    def get_universal_access(
        self, resource: str, class_name: str
    ) -> Union[str, bytes, None]:
        """Get the universal access for general resource profile."""
        profile = self.extract(resource, class_name, profile_only=True)
        return self._get_field(profile, "base", "universalAccess")

    def set_universal_access(
        self,
        resource: str,
        class_name: str,
        universal_access: str,
    ) -> Union[dict, bytes]:
        """Set the universal access for a general resource profile."""
        result = self.alter(
            resource, class_name, {"base:universal_access": universal_access}
        )
        return self._to_steps(result)

    def get_my_access(self, resource: str, class_name: str) -> Union[str, bytes, None]:
        """Get the access associated with your own general resource profile."""
        profile = self.extract(resource, class_name, profile_only=True)
        return self._get_field(profile, "base", "yourAccess")

    # ============================================================================
    # Class Administration
    # ============================================================================
    def add_resource_class(
        self, class_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new general resource class."""
        return self.add(resource=class_name, class_name="CDT", traits=traits)

    def alter_resource_class(
        self, class_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Alter an existing general resource class."""
        return self.alter(resource=class_name, class_name="CDT", traits=traits)

    def extract_resource_class(self, class_name: str) -> Union[dict, bytes]:
        """Extract the attributes of a general resource class."""
        profile = self.extract(
            resource=class_name,
            class_name="CDT",
            segments=["cdtinfo"],
            profile_only=True,
        )
        return profile["cdtinfo"]

    def delete_resource_class(self, class_name: str) -> Union[dict, bytes]:
        """Delete a general resource class."""
        return self.delete(resource=class_name, class_name="CDT")

    # ============================================================================
    # Started Task Administration
    # ============================================================================
    def add_started_task(
        self, started_task_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new started task profile."""
        return self.add(resource=started_task_name, class_name="STARTED", traits=traits)

    def alter_started_task(
        self, started_task_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Alter an existing started task profile."""
        return self.alter(
            resource=started_task_name, class_name="STARTED", traits=traits
        )

    def extract_started_task(self, started_task_name: str) -> Union[dict, bytes]:
        """Extract the attributes of a started task profile."""
        profile = self.extract(
            resource=started_task_name,
            class_name="STARTED",
            segments=["stdata"],
            profile_only=True,
        )
        return profile["stdata"]

    def delete_started_task(self, started_task_name: str) -> Union[dict, bytes]:
        """Delete a started task profile."""
        return self.delete(resource=started_task_name, class_name="STARTED")

    # ============================================================================
    # Custom Field Administration
    # ============================================================================
    def add_custom_field(
        self, custom_field_name: str, custom_field_type: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new custom field."""
        full_profile_name = f"{custom_field_type}.csdata.{custom_field_name}"
        return self.add(resource=full_profile_name, class_name="CFIELD", traits=traits)

    def alter_custom_field(
        self, custom_field_name: str, custom_field_type: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Alter an existing custom field."""
        full_profile_name = f"{custom_field_type}.csdata.{custom_field_name}"
        return self.alter(
            resource=full_profile_name, class_name="CFIELD", traits=traits
        )

    def extract_custom_field(
        self, custom_field_name: str, custom_field_type: str
    ) -> Union[dict, bytes]:
        """Extract the attributes of a custom field."""
        full_profile_name = f"{custom_field_type}.csdata.{custom_field_name}"
        profile = self.extract(
            resource=full_profile_name,
            class_name="CFIELD",
            segments=["cfdef"],
            profile_only=True,
        )
        return profile["cfdef"]

    def delete_custom_field(
        self, custom_field_name: str, custom_field_type: str
    ) -> Union[dict, bytes]:
        """Delete a custom field."""
        full_profile_name = f"{custom_field_type}.csdata.{custom_field_name}"
        return self.delete(resource=full_profile_name, class_name="CFIELD")

    # ============================================================================
    # Kerberos Realm Administration
    # ============================================================================
    def add_kerberos_realm(
        self, kerberos_realm_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new kerberos realm profile."""
        return self.add(resource=kerberos_realm_name, class_name="REALM", traits=traits)

    def alter_kerberos_realm(
        self, kerberos_realm_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Alter an existing kerberos realm profile."""
        return self.alter(
            resource=kerberos_realm_name, class_name="REALM", traits=traits
        )

    def extract_kerberos_realm(self, kerberos_realm_name: str) -> Union[dict, bytes]:
        """Extract the attributes of a kerberos realm profile."""
        profile = self.extract(
            resource=kerberos_realm_name,
            class_name="REALM",
            segments=["kerb"],
            profile_only=True,
        )
        return profile["kerb"]

    def delete_kerberos_realm(self, kerberos_realm_name: str) -> Union[dict, bytes]:
        """Delete a kerberos realm profile."""
        return self.delete(resource=kerberos_realm_name, class_name="REALM")

    # ============================================================================
    # Signed Program Administration
    # ============================================================================
    def add_signed_program(
        self, signed_program_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new signed program profile."""
        if "sigver:library" in traits:
            traits["base:member"] = traits["sigver:library"]
            del traits["sigver:library"]
        return self.add(
            resource=signed_program_name, class_name="PROGRAM", traits=traits
        )

    def alter_signed_program(
        self, signed_program_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Alter an existing signed program profile."""
        if "sigver:library" in traits:
            traits["base:member"] = traits["sigver:library"]
            del traits["sigver:library"]
        return self.alter(
            resource=signed_program_name, class_name="PROGRAM", traits=traits
        )

    def extract_signed_program(self, signed_program_name: str) -> Union[dict, bytes]:
        """Extract the attributes of a signed program profile."""
        profile = self.extract(
            resource=signed_program_name,
            class_name="PROGRAM",
            segments=["sigver"],
            profile_only=True,
        )
        profile["sigver"]["library"] = profile["base"].get("member")
        return profile["sigver"]

    def delete_signed_program(self, signed_program_name: str) -> Union[dict, bytes]:
        """Delete a signed program profile."""
        return self.delete(resource=signed_program_name, class_name="PROGRAM")

    # ============================================================================
    # APPC Session Administration
    # ============================================================================
    def add_appc_session(
        self, net_id: str, local_lu: str, partner_lu: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new APPC session profile."""
        full_profile_name = f"{net_id}.{local_lu}.{partner_lu}"
        return self.add(resource=full_profile_name, class_name="APPCLU", traits=traits)

    def alter_appc_session(
        self, net_id: str, local_lu: str, partner_lu: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Alter an existing APPC session profile."""
        full_profile_name = f"{net_id}.{local_lu}.{partner_lu}"
        return self.alter(
            resource=full_profile_name, class_name="APPCLU", traits=traits
        )

    def extract_appc_session(
        self, net_id: str, local_lu: str, partner_lu: str
    ) -> Union[dict, bytes]:
        """Extract the attributes of a APPC session profile."""
        full_profile_name = f"{net_id}.{local_lu}.{partner_lu}"
        profile = self.extract(
            resource=full_profile_name,
            class_name="APPCLU",
            segments=["session"],
            profile_only=True,
        )
        return profile["session"]

    def delete_appc_session(
        self, net_id: str, local_lu: str, partner_lu: str
    ) -> Union[dict, bytes]:
        """Delete a APPC session."""
        full_profile_name = f"{net_id}.{local_lu}.{partner_lu}"
        return self.delete(resource=full_profile_name, class_name="APPCLU")

    # ============================================================================
    # Base Functions
    # ============================================================================
    def add(
        self, resource: str, class_name: str, traits: dict = {}
    ) -> Union[dict, bytes]:
        """Create a new general resource profile."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            profile_request = ResourceRequest(resource, class_name, "set")
            self._build_xml_segments(profile_request)
            return self._make_request(profile_request)
        try:
            profile = self.extract(resource, class_name, profile_only=True)
            if self._get_field(profile, "base", "name") == resource.lower():
                raise AddOperationError(resource, class_name)
        except SecurityRequestError as exception:
            if not exception.contains_error_message(self._profile_type, "ICH13003I"):
                raise exception
        self._build_segment_trait_dictionary(traits)
        profile_request = ResourceRequest(resource, class_name, "set")
        self._build_xml_segments(profile_request)
        return self._make_request(profile_request)

    def alter(self, resource: str, class_name: str, traits: dict) -> Union[dict, bytes]:
        """Alter an existing general resource profile."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            profile_request = ResourceRequest(resource, class_name, "set")
            self._build_xml_segments(profile_request, alter=True)
            return self._make_request(profile_request, irrsmo00_precheck=True)
        try:
            profile = self.extract(resource, class_name, profile_only=True)
        except SecurityRequestError:
            raise AlterOperationError(resource, class_name)
        if not self._get_field(profile, "base", "name") == resource.lower():
            raise AlterOperationError(resource, class_name)
        self._build_segment_trait_dictionary(traits)
        profile_request = ResourceRequest(resource, class_name, "set")
        self._build_xml_segments(profile_request, alter=True)
        return self._make_request(profile_request, irrsmo00_precheck=True)

    def extract(
        self,
        resource: str,
        class_name: str,
        segments: List[str] = [],
        profile_only: bool = False,
    ) -> Union[dict, bytes]:
        """Extract a general resource profile."""
        self._build_segment_dictionary(segments)
        resource_request = ResourceRequest(resource, class_name, "listdata")
        self._build_xml_segments(resource_request, extract=True)
        result = self._extract_and_check_result(resource_request)
        if profile_only:
            return self._get_profile(result)
        return result

    def delete(self, resource: str, class_name: str) -> Union[dict, bytes]:
        """Delete a general resource profile."""
        profile_request = ResourceRequest(resource, class_name, "del")
        return self._make_request(profile_request)

    # ============================================================================
    # Private/Protected Utility Functions
    # ============================================================================
    def _format_profile(self, result: dict) -> None:
        """Format profile extract data into a dictionary."""
        messages = result["securityResult"]["resource"]["commands"][0]["messages"]
        indexes = [
            i
            for i in range(len(messages) - 1)
            if messages[i] and "CLASS      NAME" in messages[i]
        ]
        indexes.append(len(messages))
        profiles = []
        for i in range(len(indexes) - 1):
            profile = self._format_profile_generic(
                messages[indexes[i] : indexes[i + 1]]
            )
            # Post processing
            if "(g)" in profile["base"].get("name"):
                profile["base"]["generic"] = True
                profile["base"]["name"] = self._cast_from_str(
                    profile["base"].get("name")[0]
                )
            else:
                profile["base"]["generic"] = False

            if profile["base"].get("notify") == [None, "user", "to", "be", "notified"]:
                profile["base"]["notify"] = None
            profiles.append(profile)

        del result["securityResult"]["resource"]["commands"][0]["messages"]
        result["securityResult"]["resource"]["commands"][0]["profiles"] = profiles



"""Make security admin subclasses available from package root."""
from .access.access_admin import AccessAdmin
from .common.add_operation_error import AddOperationError
from .common.alter_operation_error import AlterOperationError
from .common.security_request_error import SecurityRequestError
from .common.segment_error import SegmentError
from .common.segment_trait_error import SegmentTraitError
from .connection.connection_admin import ConnectionAdmin
from .data_set.data_set_admin import DataSetAdmin
from .group.group_admin import GroupAdmin
from .resource.resource_admin import ResourceAdmin
from .setropts.setropts_admin import SetroptsAdmin
from .user.user_admin import UserAdmin



"""RACF Connection Administration."""

from typing import List, Union

from pyracf.common.security_admin import SecurityAdmin

from .connection_request import ConnectionRequest


class ConnectionAdmin(SecurityAdmin):
    """RACF Connection Administration."""

    def __init__(
        self,
        debug: bool = False,
        generate_requests_only: bool = False,
        update_existing_segment_traits: Union[dict, None] = None,
        replace_existing_segment_traits: Union[dict, None] = None,
        additional_secret_traits: Union[List[str], None] = None,
    ) -> None:
        self._valid_segment_traits = {
            "base": {
                "base:automatic_data_set_protection": "racf:adsp",
                "base:auditor": "racf:auditor",
                "base:group_authority": "racf:auth",
                "base:group": "racf:group",
                "base:group_access": "racf:grpacc",
                "base:operations": "racf:oper",
                "base:owner": "racf:owner",
                "base:resume": "racf:resume",
                "base:revoke": "racf:revoke",
                "base:special": "racf:special",
                "base:universal_access": "racf:uacc",
            }
        }
        super().__init__(
            "groupConnection",
            debug=debug,
            generate_requests_only=generate_requests_only,
            update_existing_segment_traits=update_existing_segment_traits,
            replace_existing_segment_traits=replace_existing_segment_traits,
            additional_secret_traits=additional_secret_traits,
        )

    # ============================================================================
    # Group Special
    # ============================================================================
    def give_group_special_authority(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """Give a user RACF special authority within a group."""
        result = self.connect(userid, group, {"base:special": True})
        return self._to_steps(result)

    def take_away_group_special_authority(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """Remove a user's RACF special authoritiy within a group."""
        result = self.connect(userid, group, {"base:special": False})
        return self._to_steps(result)

    # ============================================================================
    # Group Operations
    # ============================================================================
    def give_group_operations_authority(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """Give a user operations authority within a group."""
        result = self.connect(userid, group, {"base:operations": True})
        return self._to_steps(result)

    def take_away_group_operations_authority(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """Remove a user's operations authority within a group."""
        result = self.connect(userid, group, {"base:operations": False})
        return self._to_steps(result)

    # ============================================================================
    # Group Auditor
    # ============================================================================
    def give_group_auditor_authority(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """Give a user auditor authority within a group."""
        result = self.connect(userid, group, {"base:auditor": True})
        return self._to_steps(result)

    def take_away_group_auditor_authority(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """Remove a user's auditor authority within a group."""
        result = self.connect(userid, group, {"base:auditor": False})
        return self._to_steps(result)

    # ============================================================================
    # Group Access
    # ============================================================================
    def give_group_access_attribute(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """
        Automatically make group data set profiles that a user
        creates accessible to all members of the group.
        """
        result = self.connect(userid, group, {"base:group_access": True})
        return self._to_steps(result)

    def take_away_group_access_attribute(
        self, userid: str, group: str
    ) -> Union[dict, bytes]:
        """
        Don't automatically make group data set profiles that
        a user creates accessible to all members of the group.
        """
        result = self.connect(userid, group, {"base:group_access": False})
        return self._to_steps(result)

    # ============================================================================
    # Base Functions
    # ============================================================================
    def connect(self, userid: str, group: str, traits: dict = {}) -> Union[dict, bytes]:
        """Create or change a group connection."""
        self._build_segment_trait_dictionary(traits)
        connection_request = ConnectionRequest(userid, group, "set")
        self._add_traits_directly_to_request_xml_with_no_segments(
            connection_request, alter=True
        )
        return self._make_request(connection_request)

    def delete(self, userid: str, group: str) -> Union[dict, bytes]:
        """Delete a group connection."""
        connection_request = ConnectionRequest(userid, group, "del")
        self._add_traits_directly_to_request_xml_with_no_segments(connection_request)
        return self._make_request(connection_request)



"""Exception to use when the user passes bad segment-trait name(s) on an add/alter request."""


class SegmentTraitError(Exception):
    """
    Raised when a user passes a bad segment-trait combination in the traits dictionary.
    """

    def __init__(self, bad_traits: list, profile_type: str) -> None:
        self.message = "Unable to build Security Request.\n\n"
        for trait in bad_traits:
            self.message += (
                f"'{trait}' is not a known segment-trait "
                + f"combination for '{profile_type}'.\n"
            )

    def __str__(self) -> str:
        return self.message



"""Exception to use when Add operation would alter an existing profile."""


class AddOperationError(Exception):
    """
    Raised when a profile cannot be added because it already exists.
    """

    def __init__(self, profile_name: str, class_name: str) -> None:
        self.message = "Refusing to make security request to IRRSMO00."
        admin_types = ["user", "group", "dataSet"]
        if class_name not in admin_types:
            self.message += (
                "\n\nTarget profile "
                + f"'{profile_name}' already exists as a profile in the '{class_name}' class."
            )
        else:
            self.message += (
                "\n\nTarget profile "
                + f"'{profile_name}' already exists as a '{class_name}' profile."
            )

    def __str__(self) -> str:
        return self.message



"""User Administration."""

from typing import List, Union

from pyracf.common.add_operation_error import AddOperationError
from pyracf.common.alter_operation_error import AlterOperationError
from pyracf.common.security_admin import SecurityAdmin
from pyracf.common.security_request_error import SecurityRequestError

from .user_request import UserRequest


class UserAdmin(SecurityAdmin):
    """User Administration."""

    def __init__(
        self,
        debug: bool = False,
        generate_requests_only: bool = False,
        update_existing_segment_traits: Union[dict, None] = None,
        replace_existing_segment_traits: Union[dict, None] = None,
        additional_secret_traits: Union[List[str], None] = None,
    ) -> None:
        self._valid_segment_traits = {
            "base": {
                "base:automatic_data_set_protection": "racf:adsp",
                "base:auditor": "racf:auditor",
                "base:default_group_authority": "racf:auth",
                "base:security_categories": "racf:category",
                "base:class_authorizations": "racf:clauth",
                "base:installation_data": "racf:data",
                "base:default_group": "defgroup",
                "base:password_expired": "racf:expired",
                "base:group": "racf:group",
                "base:group_data_set_access": "racf:grpacc",
                "base:model_data_set": "racf:model",
                "base:name": "name",
                "base:require_operator_id_card": "racf:oidcard",
                "base:operations": "racf:oper",
                "base:owner": "racf:owner",
                "base:password": "racf:password",
                "base:passphrase": "racf:phrase",
                "base:restrict_global_access_checking": "racf:rest",
                "base:resume_date": "resumedate",
                "base:revoke_date": "revokedate",
                "base:audit_responsibility": "racf:roaudit",
                "base:security_label": "seclabel",
                "base:security_level": "racf:seclevel",
                "base:special": "racf:special",
                "base:universal_access": "racf:uacc",
                "base:audit_logging": "uaudit",
                "base:logon_allowed_days": "whendays",
                "base:logon_allowed_time": "whentime",
            },
            "cics": {
                "cics:operator_classes": "racf:opclass",
                "cics:operator_id": "opident",
                "cics:operator_priority": "opprty",
                "cics:rsl_key": "racf:rslkey",
                "cics:timeout": "timeout",
                "cics:tsl_key": "racf:tslkey",
                "cics:xrf_sign_off": "force",
            },
            "dce": {
                "dce:auto_login": "autolog",
                "dce:name": "dcename",
                "dce:home_cell": "homecell",
                "dce:home_cell_uuid": "homeuuid",
                "dce:uuid": "uuid",
            },
            "dfp": {
                "dfp:data_application": "dataappl",
                "dfp:data_class": "dataclas",
                "dfp:management_class": "mgmtclass",
                "dfp:storage_class": "storclas",
            },
            "eim": {"eim:ldap_bind_profile": "racf:ldapprof"},
            "kerb": {
                "kerb:encryption_algorithm": "racf:encrypt",
                "kerb:name": "racf:kerbname",
                "kerb:max_ticket_life": "racf:maxtktlf",
            },
            "language": {
                "language:primary": "primary",
                "language:secondary": "secondary",
            },
            "lnotes": {"lnotes:zos_short_name": "racf:sname"},
            "mfa": {
                "mfa:factor": "racf:factor",
                "mfa:active": "racf:facactv",
                "mfa:tags": "racf:factags",
                "mfa:password_fallback": "racf:mfaflbk",
                "mfa:policy": "racf:mfapolnm",
            },
            "nds": {"nds:username": "racf:uname"},
            "netview": {
                "netview:default_console": "consid",
                "netview:security_check": "secctl",
                "netview:domains": "nvdomains",
                "netview:logon_commands": "ic",
                "netview:recieve_unsolicited_messages": "msgrec",
                "netview:graphic_monitor_facility_admin": "racf:ngmfadmn",
                "netview:view_span": "gmfadmin",
                "netview:operator_scope_classes": "racf:opclass",
            },
            "omvs": {
                "omvs:max_address_space_size": "assize",
                "omvs:auto_uid": "racf:autouid",
                "omvs:max_cpu_time": "cputime",
                "omvs:max_files_per_process": "filemax",
                "omvs:home_directory": "home",
                "omvs:max_non_shared_memory": "memlim",
                "omvs:max_file_mapping_pages": "mmaparea",
                "omvs:max_processes": "procmax",
                "omvs:default_shell": "pgm",
                "omvs:shared": "racf:shared",
                "omvs:max_shared_memory": "shmemmax",
                "omvs:max_threads": "threads",
                "omvs:uid": "uid",
            },
            "operparm": {
                "operparm:alternate_console_group": "altgrp",
                "operparm:recieve_automated_messages": "auto",
                "operparm:command_target_systems": "cmdsys",
                "operparm:recieve_delete_operator_messages": "dom",
                "operparm:recieve_hardcopy_messages": "hc",
                "operparm:recieve_internal_console_messages": "intid",
                "operparm:console_searching_key": "key",
                "operparm:message_level": "racf:level",
                "operparm:log_command_responses": "logcmd",
                "operparm:message_format": "mform",
                "operparm:migration_id": "migid",
                "operparm:events_to_monitor": "mon",
                "operparm:message_scope": "racf:mscope",
                "operparm:operator_command_authority": "auth",
                "operparm:recieve_routing_codes": "routcode",
                "operparm:message_queue_storage": "storage",
                "operparm:recieve_undelivered_messages": "ud",
                "operparm:recieve_messages_from_unknown_console_ids": "unkids",
            },
            "ovm": {
                "ovm:file_system_root": "racf:fsroot",
                "ovm:home_directory": "racf:vhome",
                "ovm:default_shell": "racf:vprogram",
                "ovm:uid": "racf:vuid",
            },
            "proxy": {
                "proxy:bind_distinguished_name": "racf:binddn",
                "proxy:bind_password": "racf:bindpw",
                "proxy:ldap_host": "racf:ldaphost",
            },
            "tso": {
                "tso:account_number": "acctnum",
                "tso:logon_command": "command",
                "tso:sysout_destination_id": "dest",
                "tso:hold_class": "holdclass",
                "tso:job_class": "jobclass",
                "tso:max_region_size": "maxsize",
                "tso:message_class": "msgclass",
                "tso:logon_procedure": "proc",
                "tso:security_label": "seclabel",
                "tso:default_region_size": "size",
                "tso:sysout_class": "sysclass",
                "tso:data_set_allocation_unit": "unit",
                "tso:user_data": "userdata",
            },
            "workattr": {
                "workattr:account_number": "waaccnt",
                "workattr:sysout_address_1": "waaddr1",
                "workattr:sysout_address_2": "waaddr2",
                "workattr:sysout_address_3": "waaddr3",
                "workattr:sysout_address_4": "waaddr4",
                "workattr:sysout_building": "wabldg",
                "workattr:sysout_department": "wadept",
                "workattr:sysout_user": "waname",
                "workattr:sysout_room": "waroom",
                "workattr:sysout_email": "waemail",
            },
        }
        self._extracted_key_value_pair_segment_traits_map = {
            "omvs": {
                "home": "homeDirectory",
                "program": "defaultShell",
                "cputimemax": "maxCpuTime",
                "assizemax": "maxAddressSpaceSize",
                "fileprocmax": "maxFilesPerProcess",
                "procusermax": "maxProcesses",
                "threadsmax": "maxThreads",
                "mmapareamax": "maxFileMappingPages",
                "memlimit": "maxNonSharedMemory",
                "shmemmax": "maxSharedMemory",
            },
            "tso": {
                "acctnum": "accountNumber",
                "holdclass": "holdClass",
                "jobclass": "jobClass",
                "msgclass": "messageClass",
                "proc": "logonProcedure",
                "size": "defaultRegionSize",
                "maxsize": "maxRegionSize",
                "sysoutclass": "sysoutClass",
                "unit": "dataSetAllocationUnit",
                "userdata": "userData",
                "command": "logonCommand",
            },
        }
        self._case_sensitive_extracted_values = ["homeDirectory", "defaultShell"]
        super().__init__(
            "user",
            debug=debug,
            generate_requests_only=generate_requests_only,
            update_existing_segment_traits=update_existing_segment_traits,
            replace_existing_segment_traits=replace_existing_segment_traits,
            additional_secret_traits=additional_secret_traits,
        )

    # ============================================================================
    # Special Authority
    # ============================================================================
    def has_special_authority(self, userid: str) -> Union[bool, bytes]:
        """Check if a user has RACF special authority."""
        profile = self.extract(userid, profile_only=True)
        return "special" in profile["base"]["attributes"]

    def give_special_authority(self, userid: str) -> Union[dict, bytes]:
        """Give a user RACF special authority."""
        result = self.alter(userid, traits={"base:special": True})
        return self._to_steps(result)

    def take_away_special_authority(
        self,
        userid: str,
    ) -> Union[dict, bytes]:
        """Remove a user's RACF special authority."""
        result = self.alter(userid, traits={"base:special": False})
        return self._to_steps(result)

    # ============================================================================
    # Operations Authority
    # ============================================================================
    def has_operations_authority(self, userid: str) -> Union[bool, bytes]:
        """Check if a user has operations authority."""
        profile = self.extract(userid, profile_only=True)
        return "operations" in profile["base"]["attributes"]

    def give_operations_authority(self, userid: str) -> Union[dict, bytes]:
        """Give a user operations authority."""
        result = self.alter(userid, traits={"base:operations": True})
        return self._to_steps(result)

    def take_away_operations_authority(self, userid: str) -> Union[dict, bytes]:
        """Remove a user's operations authority."""
        result = self.alter(userid, traits={"base:operations": False})
        return self._to_steps(result)

    # ============================================================================
    # Auditor Authority
    # ============================================================================
    def has_auditor_authority(self, userid: str) -> Union[bool, bytes]:
        """Check if a user has auditor authority."""
        profile = self.extract(userid, profile_only=True)
        return "auditor" in profile["base"]["attributes"]

    def give_auditor_authority(self, userid: str) -> Union[dict, bytes]:
        """Give a user auditor authority."""
        result = self.alter(userid, traits={"base:auditor": True})
        return self._to_steps(result)

    def take_away_auditor_authority(self, userid: str) -> Union[dict, bytes]:
        """Remove a user's auditor authority."""
        result = self.alter(userid, traits={"base:auditor": False})
        return self._to_steps(result)

    # ============================================================================
    # Password
    # ============================================================================
    def set_password(
        self,
        userid: str,
        password: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's password."""
        result = self.alter(userid, traits={"base:password": password})
        return self._to_steps(result)

    # ============================================================================
    # Passphrase
    # ============================================================================
    def set_passphrase(
        self,
        userid: str,
        passphrase: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's passphrase."""
        result = self.alter(userid, traits={"base:passphrase": passphrase})
        return self._to_steps(result)

    # ============================================================================
    # Class Authorizations
    # ============================================================================
    def get_class_authorizations(self, userid: str) -> Union[List[str], bytes]:
        """Get a user's class authorizations."""
        profile = self.extract(userid, profile_only=True)
        return self._get_field(profile, "base", "classAuthorizations")

    def set_class_authorizations(
        self, userid: str, class_authorizations: List[str]
    ) -> Union[dict, bytes]:
        """
        Set a user's class authorizations.
        removes the user's current class authorizations and then recreates
        the class authorizations list using the list that the user provides.
        """
        delete_result = self.remove_all_class_authorizations(userid)
        add_result = self.add_class_authorizations(userid, class_authorizations)
        return self._to_steps([delete_result, add_result])

    def add_class_authorizations(
        self, userid: str, class_authorizations: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Add one or more classes to a user's class authorizations."""
        result = self.alter(
            userid, traits={"add:base:class_authorizations": class_authorizations}
        )
        return self._to_steps(result)

    def remove_class_authorizations(
        self, userid: str, class_authorizations: Union[str, List[str]]
    ) -> Union[dict, bytes]:
        """Remove one or more classes from a user's class authorizations."""
        result = self.alter(
            userid, traits={"remove:base:class_authorizations": class_authorizations}
        )
        return self._to_steps(result)

    def remove_all_class_authorizations(self, userid: str) -> Union[dict, bool, bytes]:
        """Remove all classes from a users class authorizations."""
        current_class_authorizations = self.get_class_authorizations(userid)
        if not current_class_authorizations:
            return False
        return self.remove_class_authorizations(userid, current_class_authorizations)

    # ============================================================================
    # Revoke Date
    # ============================================================================
    def get_revoke_date(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's revoke date."""
        profile = self.extract(userid, profile_only=True)
        return self._get_field(profile, "base", "revokeDate", string=True)

    def set_revoke_date(
        self, userid: str, revoke_date: Union[str, bool]
    ) -> Union[dict, bytes]:
        """Set a user's revoke date."""
        result = self.alter(userid, traits={"base:revoke_date": revoke_date})
        return self._to_steps(result)

    # ============================================================================
    # Resume Date
    # ============================================================================
    def get_resume_date(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's resume date."""
        profile = self.extract(userid, profile_only=True)
        return self._get_field(profile, "base", "resumeDate", string=True)

    def set_resume_date(
        self, userid: str, resume_date: Union[str, bool]
    ) -> Union[dict, bytes]:
        """Set a user's resume date."""
        result = self.alter(userid, traits={"base:resume_date": resume_date})
        return self._to_steps(result)

    # ============================================================================
    # Owner
    # ============================================================================
    def get_owner(self, userid: str) -> Union[str, bytes]:
        """Get a user's owner."""
        profile = self.extract(userid, profile_only=True)
        return self._get_field(profile, "base", "owner", string=True)

    def set_owner(self, userid: str, owner: str) -> Union[dict, bytes]:
        """Set a user's owner."""
        result = self.alter(userid, traits={"base:owner": owner})
        return self._to_steps(result)

    # ============================================================================
    # Name
    # ============================================================================
    def get_name(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's name."""
        profile = self.extract(userid, profile_only=True)
        return self._get_field(profile, "base", "name", string=True)

    def set_name(self, userid: str, name: Union[str, bool]) -> Union[dict, bytes]:
        """Set a user's name."""
        result = self.alter(userid, traits={"base:name": name})
        return self._to_steps(result)

    # ============================================================================
    # OMVS UID
    # ============================================================================
    def get_omvs_uid(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS UID."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "uid")

    def set_omvs_uid(self, userid: str, uid: Union[int, bool]) -> Union[dict, bytes]:
        """Set a user's OMVS UID."""
        result = self.alter(userid, traits={"omvs:uid": uid})
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max Address Space Size
    # ============================================================================
    def get_omvs_max_address_space_size(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS max address space size."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxAddressSpaceSize")

    def set_omvs_max_address_space_size(
        self,
        userid: str,
        max_address_space_size: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max address space size."""
        result = self.alter(
            userid, traits={"omvs:max_address_space_size": max_address_space_size}
        )
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max CPU Time
    # ============================================================================
    def get_omvs_max_cpu_time(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS max cpu time."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxCpuTime")

    def set_omvs_max_cpu_time(
        self,
        userid: str,
        max_cpu_time: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max cpu time."""
        result = self.alter(userid, traits={"omvs:max_cpu_time": max_cpu_time})
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max Files Per Process
    # ============================================================================
    def get_omvs_max_files_per_process(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS max files per process."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxFilesPerProcess")

    def set_omvs_max_files_per_process(
        self,
        userid: str,
        max_files_per_process: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max files per process."""
        result = self.alter(
            userid, traits={"omvs:max_files_per_process": max_files_per_process}
        )
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max Non-Shared Memory
    # ============================================================================
    def get_omvs_max_non_shared_memory(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's OMVS max non-shared memory."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxNonSharedMemory", string=True)

    def set_omvs_max_non_shared_memory(
        self,
        userid: str,
        max_non_shared_memory: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max non-shared memory."""
        result = self.alter(
            userid, traits={"omvs:max_non_shared_memory": max_non_shared_memory}
        )
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max File Mapping Pages
    # ============================================================================
    def get_omvs_max_file_mapping_pages(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS max file mapping pages."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxFileMappingPages")

    def set_omvs_max_file_mapping_pages(
        self,
        userid: str,
        max_file_mapping_pages: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max file mapping pages."""
        result = self.alter(
            userid, traits={"omvs:max_file_mapping_pages": max_file_mapping_pages}
        )
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max Processes
    # ============================================================================
    def get_omvs_max_processes(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS max processes."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxProcesses")

    def set_omvs_max_processes(
        self,
        userid: str,
        max_processes: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max processes."""
        result = self.alter(userid, traits={"omvs:max_processes": max_processes})
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max Shared Memory
    # ============================================================================
    def get_omvs_max_shared_memory(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's OMVS max shared memory."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxSharedMemory", string=True)

    def set_omvs_max_shared_memory(
        self,
        userid: str,
        max_shared_memory: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max shared memory."""
        result = self.alter(
            userid, traits={"omvs:max_shared_memory": max_shared_memory}
        )
        return self._to_steps(result)

    # ============================================================================
    # OMVS Max Threads
    # ============================================================================
    def get_omvs_max_threads(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's OMVS max threads."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "maxThreads")

    def set_omvs_max_threads(
        self,
        userid: str,
        max_threads: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS max threads."""
        result = self.alter(userid, traits={"omvs:max_threads": max_threads})
        return self._to_steps(result)

    # ============================================================================
    # OMVS Home Directory
    # ============================================================================
    def get_omvs_home_directory(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's OMVS home directory."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "homeDirectory", string=True)

    def set_omvs_home_directory(
        self,
        userid: str,
        home_directory: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS home directory."""
        result = self.alter(userid, traits={"omvs:home_directory": home_directory})
        return self._to_steps(result)

    # ============================================================================
    # OMVS Default Shell
    # ============================================================================
    def get_omvs_default_shell(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's OMVS default shell."""
        profile = self.extract(userid, segments=["omvs"], profile_only=True)
        return self._get_field(profile, "omvs", "defaultShell", string=True)

    def set_omvs_default_shell(
        self,
        userid: str,
        default_shell: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's OMVS default shell."""
        result = self.alter(userid, traits={"omvs:default_shell": default_shell})
        return self._to_steps(result)

    # ============================================================================
    # TSO Account Number
    # ============================================================================
    def get_tso_account_number(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO account number."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "accountNumber", string=True)

    def set_tso_account_number(
        self,
        userid: str,
        account_number: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO account number."""
        result = self.alter(userid, traits={"tso:account_number": account_number})
        return self._to_steps(result)

    # ============================================================================
    # TSO Logon Command
    # ============================================================================
    def get_tso_logon_command(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO logon command."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "logonCommand", string=True)

    def set_tso_logon_command(
        self,
        userid: str,
        logon_command: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO logon command."""
        result = self.alter(userid, traits={"tso:logon_command": logon_command})
        return self._to_steps(result)

    # ============================================================================
    # TSO Hold Class
    # ============================================================================
    def get_tso_hold_class(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO hold class."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "holdClass", string=True)

    def set_tso_hold_class(
        self,
        userid: str,
        hold_class: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO hold class."""
        result = self.alter(userid, traits={"tso:hold_class": hold_class})
        return self._to_steps(result)

    # ============================================================================
    # TSO Max Region Size
    # ============================================================================
    def get_tso_max_region_size(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's TSO max region size."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "maxRegionSize")

    def set_tso_max_region_size(
        self,
        userid: str,
        max_region_size: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO max region size."""
        result = self.alter(userid, traits={"tso:max_region_size": max_region_size})
        return self._to_steps(result)

    # ============================================================================
    # TSO Message Class
    # ============================================================================
    def get_tso_message_class(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO message class."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "messageClass", string=True)

    def set_tso_message_class(
        self,
        userid: str,
        message_class: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO message class."""
        result = self.alter(userid, traits={"tso:message_class": message_class})
        return self._to_steps(result)

    # ============================================================================
    # TSO Logon Procedure
    # ============================================================================
    def get_tso_logon_procedure(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO logon procedure."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "logonProcedure", string=True)

    def set_tso_logon_procedure(
        self,
        userid: str,
        logon_procedure: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO logon procedure."""
        result = self.alter(userid, traits={"tso:logon_procedure": logon_procedure})
        return self._to_steps(result)

    # ============================================================================
    # TSO Default Region Size
    # ============================================================================
    def get_tso_default_region_size(self, userid: str) -> Union[int, None, bytes]:
        """Get a user's TSO default region size."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "defaultRegionSize")

    def set_tso_default_region_size(
        self,
        userid: str,
        default_region_size: Union[int, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO default region size."""
        result = self.alter(
            userid, traits={"tso:default_region_size": default_region_size}
        )
        return self._to_steps(result)

    # ============================================================================
    # TSO Sysout Class
    # ============================================================================
    def get_tso_sysout_class(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO sysout class."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "sysoutClass", string=True)

    def set_tso_sysout_class(
        self,
        userid: str,
        sysout_class: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO sysout class."""
        result = self.alter(userid, traits={"tso:sysout_class": sysout_class})
        return self._to_steps(result)

    # ============================================================================
    # TSO User Data
    # ============================================================================
    def get_tso_user_data(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO user data."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "userData", string=True)

    def set_tso_user_data(
        self,
        userid: str,
        user_data: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO user data."""
        result = self.alter(userid, traits={"tso:user_data": user_data})
        return self._to_steps(result)

    # ============================================================================
    # TSO Data Set Allocation Unit
    # ============================================================================
    def get_tso_data_set_allocation_unit(self, userid: str) -> Union[str, None, bytes]:
        """Get a user's TSO data set allocation unit."""
        profile = self.extract(userid, segments=["tso"], profile_only=True)
        return self._get_field(profile, "tso", "dataSetAllocationUnit", string=True)

    def set_tso_data_set_allocation_unit(
        self,
        userid: str,
        data_set_allocation_unit: Union[str, bool],
    ) -> Union[dict, bytes]:
        """Set a user's TSO data set allocation unit."""
        result = self.alter(
            userid, traits={"tso:data_set_allocation_unit": data_set_allocation_unit}
        )
        return self._to_steps(result)

    # ============================================================================
    # Base Functions
    # ============================================================================
    def add(self, userid: str, traits: dict = {}) -> Union[dict, bytes]:
        """Create a new user."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            user_request = UserRequest(userid, "set")
            self._build_xml_segments(user_request)
            return self._make_request(user_request)
        try:
            self.extract(userid)
        except SecurityRequestError as exception:
            if not exception.contains_error_message(self._profile_type, "ICH30001I"):
                raise exception
            self._build_segment_trait_dictionary(traits)
            user_request = UserRequest(userid, "set")
            self._build_xml_segments(user_request)
            return self._make_request(user_request)
        raise AddOperationError(userid, self._profile_type)

    def alter(self, userid: str, traits: dict) -> Union[dict, bytes]:
        """Alter an existing user."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            user_request = UserRequest(userid, "set")
            self._build_xml_segments(user_request, alter=True)
            return self._make_request(user_request, irrsmo00_precheck=True)
        try:
            self.extract(userid)
        except SecurityRequestError:
            raise AlterOperationError(userid, self._profile_type)
        self._build_segment_trait_dictionary(traits)
        user_request = UserRequest(userid, "set")
        self._build_xml_segments(user_request, alter=True)
        return self._make_request(user_request, irrsmo00_precheck=True)

    def extract(
        self, userid: str, segments: List[str] = [], profile_only: bool = False
    ) -> Union[dict, bytes]:
        """Extract a user's profile."""
        self._build_segment_dictionary(segments)
        user_request = UserRequest(userid, "listdata")
        self._build_xml_segments(user_request, extract=True)
        result = self._extract_and_check_result(user_request)
        if profile_only:
            return self._get_profile(result)
        return result

    def delete(self, userid: str) -> Union[dict, bytes]:
        """Delete a user."""
        user_request = UserRequest(userid, "del")
        return self._make_request(user_request)

    # ============================================================================
    # Private/Protected Utility Functions
    # ============================================================================
    def _format_profile(self, result: dict) -> None:
        """Format profile extract data into a dictionary."""
        messages = result["securityResult"]["user"]["commands"][0]["messages"]
        userid = result["securityResult"]["user"]["name"]
        exclude_command_audit_trail_string = f"Command Audit Trail for USER {userid}"
        if exclude_command_audit_trail_string in messages:
            messages = messages[: messages.index(exclude_command_audit_trail_string)]
        indexes = [
            i
            for i in range(len(messages) - 1)
            if messages[i] and "USER=" in messages[i]
        ]
        indexes.append(len(messages))
        profiles = []
        for i in range(len(indexes) - 1):
            profile = self._format_profile_generic(
                messages[indexes[i] : indexes[i + 1]]
            )
            profiles.append(profile)

        # Post processing
        del result["securityResult"]["user"]["commands"][0]["messages"]
        result["securityResult"]["user"]["commands"][0]["profiles"] = profiles



"""User Administration Request Builder."""

from pyracf.common.security_request import SecurityRequest


class UserRequest(SecurityRequest):
    """User Administration Request Builder."""

    def __init__(self, userid: str, operation: str) -> None:
        super().__init__()
        self._security_definition.tag = "user"
        self._security_definition.attrib = {
            "name": userid,
            "operation": operation,
            "requestid": "UserRequest",
        }



"""Logging for pyRACF."""

import inspect
import json
import os
import re
from typing import Union


class Logger:
    """Logging for pyRACF."""

    __ansi_reset = "\033[0m"
    __ansi_gray = "\033[2m"
    __ansi_green = "\033[32m"
    __ansi_blue = "\033[34m"
    __ansi_orange = "\033[38;5;214m"
    __ansi_cyan = "\033[96m"
    __ansi_purple_background = "\033[1;45m"

    def __gray(self, string: str) -> str:
        """Make contents of string gray."""
        return self.__colorize_string(self.__ansi_gray, string)

    def __green(self, string: str) -> str:
        """Make contents of string green."""
        return self.__colorize_string(self.__ansi_green, string)

    def __blue(self, string: str) -> str:
        """Make contents of string blue."""
        return self.__colorize_string(self.__ansi_blue, string)

    def __orange(self, string: str) -> str:
        """Make contents of string orange."""
        return self.__colorize_string(self.__ansi_orange, string)

    def __cyan(self, string: str) -> str:
        """Make contents of string cyan."""
        return self.__colorize_string(self.__ansi_cyan, string)

    def __purple_background(self, string: str) -> str:
        """Make contents of string magenta."""
        return self.__colorize_string(self.__ansi_purple_background, string)

    def __colorize_string(self, ansi_color: str, string: str) -> str:
        return f"{ansi_color}{string}{self.__ansi_reset}"

    def log_dictionary(
        self,
        header_message: str,
        dictionary: dict,
        secret_traits: dict = {},
    ) -> None:
        """JSONify and colorize a dictionary and log it to the console."""
        if secret_traits:
            dictionary = self.__redact_request_dictionary(dictionary, secret_traits)
        dictionary_json = json.dumps(dictionary, indent=2)
        colorized_dictionary_json = self.__colorize_json(dictionary_json)
        self.log_debug(header_message, colorized_dictionary_json)

    def log_xml(
        self,
        header_message: str,
        xml_string: Union[str, bytes],
        secret_traits: dict = {},
    ) -> None:
        """Indent and colorize XML string and log it to the console."""
        if isinstance(xml_string, bytes):
            xml_string = xml_string.decode(encoding="utf-8")
        if secret_traits:
            xml_string = self.redact_request_xml(xml_string, secret_traits)
        indented_xml_string = self.__indent_xml(xml_string)
        colorized_indented_xml_string = self.__colorize_xml(indented_xml_string)
        self.log_debug(header_message, colorized_indented_xml_string)

    def log_debug(self, header_message: str, message: str) -> None:
        """Log function to use for debug logging."""
        stack = list(reversed(inspect.stack()))
        admin_class_calls = [
            (os.path.basename(stack[i].filename), stack[i].function)
            for i in range(len(stack) - 1)
            if os.path.basename(stack[i + 1].filename).split("_")[-1] == "admin.py"
        ][1:]
        user_call = admin_class_calls[0]
        class_tokens = [
            token.title() for token in user_call[0].split(".")[0].split("_")
        ]
        admin_class = "".join(class_tokens)
        function_called_by_user = f"{admin_class}.{user_call[1]}()"
        header = (
            f"{self.__purple_background(' '*79)}\n"
            + f"{self.__purple_background('[pyRACF:Debug]'.center(79))}\n"
            + f"{self.__purple_background(header_message.center(79))}\n"
            + f"{self.__purple_background(function_called_by_user.center(79))}\n"
            + f"{self.__purple_background(' '*79)}\n"
        )
        print(f"{header}\n{message}")

    def __redact_request_dictionary(
        self,
        dictionary: dict,
        secret_traits: dict,
    ) -> dict:
        """
        Redact a list of specified secret traits in a request dictionary.
        While this flow is technically used in extract functions, there are no traits to redact.
        """
        for trait in secret_traits:
            segment = trait.split(":")[0]
            if not dictionary.get(segment, {}).get(trait, {}).get("value", None):
                continue
            dictionary[segment][trait]["value"] = "********"
        return dictionary

    def __redact_string(
        self,
        input_string: str,
        start_ind: int,
        end_pattern: str,
    ):
        """
        Redacts characters in a string between a starting index and ending pattern.
        Replaces the identified characters with '********' regardless of the original length.
        """
        pre_keyword = input_string[:start_ind]
        post_keyword = end_pattern.join(input_string[start_ind:].split(end_pattern)[1:])
        return pre_keyword + "********" + end_pattern + post_keyword

    def redact_request_xml(
        self,
        xml_string: Union[str, bytes],
        secret_traits: dict,
    ) -> Union[str, bytes]:
        """
        Redact a list of specific secret traits in a request xml string or bytes object.
        Based the following xml pattern:
            '<xmltag attribute="any">xml value</xmltag>'
        This function also accounts for any number of arbitrary xml attributes.
        """
        is_bytes = False
        if isinstance(xml_string, bytes):
            is_bytes = True
            xml_string = xml_string.decode("utf-8")
        for xml_key in secret_traits.values():
            match = re.search(rf"\<{xml_key}+[^>]*\>", xml_string)
            if not match:
                continue
            # Delete operation has no end tag and and redaction should not be attempted.
            #
            # Redact this:
            # <tag operation="set">secret</tag>
            #
            # Don't try to redact this:
            # <tag operation="del" />
            if f"</{xml_key}>" not in xml_string:
                continue
            xml_string = self.__redact_string(xml_string, match.end(), f"</{xml_key}")
        if is_bytes:
            xml_string = xml_string.encode("utf-8")
        return xml_string

    def redact_result_xml(
        self,
        xml_string: str,
        secret_traits: dict,
    ) -> str:
        """
        Redacts a list of specific secret traits in a result xml string.
        Based on the following RACF command pattern:
            'TRAIT (value)'
        This function also accounts for varied amounts of whitespace in the pattern.
        """
        for xml_key in secret_traits.values():
            racf_key = xml_key.split(":")[1] if ":" in xml_key else xml_key
            match = re.search(rf"{racf_key.upper()} +\(", xml_string)
            if not match:
                continue
            xml_string = self.__redact_string(xml_string, match.end(), ")")
        return xml_string

    def __colorize_json(self, json_text: str) -> str:
        updated_json_text = ""
        json_lines = json_text.splitlines()
        for line in json_lines:
            tokens = line.split('"')
            if len(tokens) < 3:
                pass
            elif len(tokens[2]) == 0 or len(tokens[2]) == 1:
                line = self.__colorize_json_value(tokens)
            elif tokens[2][0] == ":":
                key = f'"{tokens[1]}"'
                first_half = f"{tokens[0]}{self.__cyan(key)}"
                second_half = self.__colorize_json_value(tokens[2:])
                line = first_half + second_half
            updated_json_text += f"{line}\n"
        return updated_json_text

    def __colorize_json_value(self, tokens: dict) -> str:
        if len(tokens) != 1:
            value = f'"{tokens[1]}"'
            second_half = f"{tokens[0]}{self.__orange(value)}{''.join(tokens[2:])}"
        else:
            skip_values = ["{", "[", "{}", "[]"]
            blue_values = ["null", "true", "false"]
            if "," in tokens[0]:
                value = tokens[0][2:-1]
                has_comma = True
            else:
                value = tokens[0][2:]
                has_comma = False
            if value in skip_values:
                pass
            elif value in blue_values:
                # null and booleans
                value = f"{self.__blue(value)}"
            else:
                # numbers
                value = f"{self.__green(value)}"
            second_half = f": {value}"
            if has_comma:
                second_half += ","
        return second_half

    def __colorize_xml(self, xml_text: str) -> str:
        updated_xml_text = ""
        xml_lines = xml_text.splitlines()
        for line in xml_lines:
            tokens = line.split("<")
            if ">" in tokens[1] and not tokens[1][-1] == ">":
                tag_tokens = tokens[1].split(">")[0].split()
            else:
                tag_tokens = tokens[1].split()
            tag_name = tag_tokens[0]
            tag_start = "<"
            tag_end = ">"
            if tag_name[0] == "?":
                tag_start = "<?"
                tag_end = "?>"
                tag_name = tag_name[1:]
            elif tag_name[0] == "/":
                tag_start = "</"
                tag_name = tag_name[1:]
            if ">" in tag_name:
                tag_name = tag_name.split(">")[0]
                tag_tokens = []
            attributes = " ".join(tag_tokens[1:])[: -len(tag_end)]
            if len(attributes) > 0:
                if attributes.count('"') % 2:
                    attributes = attributes + '"'
                if attributes[-1] == "/":
                    attributes = attributes[:-1]
                    tag_end = "/>"
            start_tag = (
                f"{self.__gray(tag_start)}"
                + f"{self.__blue(tag_name)}"
                + f"{self.__colorize_xml_attributes(attributes)}"
                + f"{self.__gray(tag_end)}"
            )
            data = ""
            end_tag = ""
            if len(tokens) > 2:
                data = "".join(tokens[1].split(">")[1:])
                end_tag = (
                    f"{self.__gray('</')}"
                    + f"{self.__blue(tokens[-1][1:-1])}"
                    + f"{self.__gray('>')}"
                )
            line = tokens[0] + start_tag + data + end_tag
            updated_xml_text += f"{line}\n"
        return updated_xml_text

    def __colorize_xml_attributes(self, xml_attributes: str) -> str:
        attribute_tokens = xml_attributes.split("=")
        updated_xml_attributes = " "
        if len(attribute_tokens) != 1:
            for token in attribute_tokens:
                subtokens = token.split()
                for subtoken in subtokens:
                    if '"' in subtoken:
                        updated_xml_attributes += f"{self.__orange(subtoken)} "
                    else:
                        updated_xml_attributes += f"{self.__cyan(subtoken)}="
        return updated_xml_attributes.rstrip()

    def __indent_xml(self, minified_xml: str) -> str:
        """Used to indent XML for readability in debug logging."""
        indent_level = 0
        indented_xml = ""
        xml_tokens = minified_xml.split("><")
        for i in range(1, len(xml_tokens)):
            previous_line = xml_tokens[i - 1]
            current_line = xml_tokens[i]
            if previous_line[:5] == "<?xml":
                indented_xml += f"{previous_line}>\n"
            elif i == 1:
                indented_xml += f"{previous_line}>\n"
                indent_level += 1
            elif (
                "</" not in previous_line
                and previous_line[0] != "/"
                and previous_line[-1] != "/"
            ):
                indent_level += 1
            if current_line[0] == "/":
                indent_level -= 1
            current_line = f"<{current_line}>"
            indented_xml += f"{'  ' * indent_level}{current_line}\n"
        return indented_xml[:-2]



"""Exception to use when the user passes bad segment name(s) on an extract request."""


class SegmentError(Exception):
    """
    Raised when a user passes a bad segment name on an extract request.
    """

    def __init__(self, bad_segments: list, profile_type: str) -> None:
        self.message = "Unable to build Security Request.\n\n"
        for segment in bad_segments:
            self.message += (
                f"'{segment}' is not a known segment for '{profile_type}'.\n"
            )

    def __str__(self) -> str:
        return self.message



"""Set RCAF Options Administration Request Builder."""

from pyracf.common.security_request import SecurityRequest


class SetroptsRequest(SecurityRequest):
    """Set RCAF Options Administration Request Builder."""

    def __init__(self) -> None:
        super().__init__()
        self._security_definition.tag = "systemsettings"
        self._security_definition.attrib = {
            "operation": "set",
            "requestid": "SetroptsRequest",
        }



"""Data Set Profile Administration Request Builder."""

from typing import Union

from pyracf.common.security_request import SecurityRequest


class DataSetRequest(SecurityRequest):
    """Data Set Profile Administration Request Builder."""

    def __init__(
        self, data_set: str, operation: str, volume: Union[str, bool], generic: bool
    ) -> None:
        super().__init__()
        self._security_definition.tag = "dataset"
        (volume, generic) = self._get_volume_and_generic_security_definition_values(
            volume, generic
        )
        self._security_definition.attrib.update(
            {
                "name": data_set,
                "operation": operation,
                "generic": generic,
                "volume": volume,
                "requestid": "DatasetRequest",
            }
        )
        if volume == "":
            del self._security_definition.attrib["volume"]



"""RACF Data Set Profile Administration."""

from typing import List, Union

from pyracf.common.add_operation_error import AddOperationError
from pyracf.common.alter_operation_error import AlterOperationError
from pyracf.common.security_admin import SecurityAdmin
from pyracf.common.security_request_error import SecurityRequestError

from .data_set_request import DataSetRequest


class DataSetAdmin(SecurityAdmin):
    """RACF Data Set Profile Administration."""

    def __init__(
        self,
        debug: bool = False,
        generate_requests_only: bool = False,
        update_existing_segment_traits: Union[dict, None] = None,
        replace_existing_segment_traits: Union[dict, None] = None,
        additional_secret_traits: Union[List[str], None] = None,
    ) -> None:
        self._valid_segment_traits = {
            "base": {
                "base:alter_volume": "racf:altvol",
                "base:audit_alter": "racf:audaltr",
                "base:audit_control": "racf:audcntl",
                "base:audit_none": "racf:audnone",
                "base:audit_read": "racf:audread",
                "base:audit_update": "racf:audupdt",
                "base:security_categories": "racf:category",
                "base:installation_data": "racf:data",
                "base:erase_data_sets_on_delete": "racf:erase",
                "base:model_profile_class": "racf:fclass",
                "base:model_profile_generic": "racf:fgeneric",
                "base:tape_data_set_file_sequence_number": "racf:fileseq",
                "base:model_profile": "racf:from",
                "base:model_profile_volume": "racf:fvolume",
                "base:global_audit_alter": "racf:gaudaltr",
                "base:global_audit_control": "racf:gaudcntl",
                "base:global_audit_none": "racf:gaudnone",
                "base:global_audit_read": "racf:gaudread",
                "base:global_audit_update": "racf:gaudupdt",
                "base:level": "racf:level",
                "base:data_set_model_profile": "racf:model",
                "base:notify_userid": "racf:notify",
                "base:owner": "racf:owner",
                "base:tape_data_set_security_retention_period": "racf:retpd",
                "base:security_label": "racf:seclabel",
                "base:security_level": "racf:seclevel",
                "base:generic_not_allowed": "racf:set",
                "base:generic_allowed": "racf:setonly",
                "base:use_tape_data_set_profile": "racf:tape",
                "base:universal_access": "racf:uacc",
                "base:data_set_allocation_unit": "racf:unit",
                "base:volumes": "racf:volume",
                "base:warn_on_insufficient_access": "racf:warning",
            },
            "dfp": {"dfp:owner": "racf:resowner", "dfp:ckds_data_key": "racf:datakey"},
            "tme": {"tme:roles": "racf:roles"},
        }
        super().__init__(
            "dataSet",
            debug=debug,
            generate_requests_only=generate_requests_only,
            update_existing_segment_traits=update_existing_segment_traits,
            replace_existing_segment_traits=replace_existing_segment_traits,
            additional_secret_traits=additional_secret_traits,
        )
        self._valid_segment_traits["base"].update(
            self._common_base_traits_data_set_generic
        )
        del self._valid_segment_traits["base"]["base:generic"]

    # ============================================================================
    # Access
    # ============================================================================
    def get_universal_access(self, data_set: str) -> Union[str, bytes, None]:
        """Get universal access for data set profile."""
        profile = self.extract(data_set, profile_only=True)
        return self._get_field(profile, "base", "universalAccess")

    def set_universal_access(
        self, data_set: str, universal_acccess: str
    ) -> Union[dict, bytes]:
        """Set the universal access for a data set profile."""
        result = self.alter(data_set, {"base:universal_access": universal_acccess})
        return self._to_steps(result)

    def get_my_access(self, data_set: str) -> Union[str, bytes, None]:
        """Get the access associated with your own data set profile."""
        profile = self.extract(data_set, profile_only=True)
        return self._get_field(profile, "base", "yourAccess")

    # ============================================================================
    # Base Functions
    # ============================================================================
    def add(
        self,
        data_set: str,
        traits: dict = {},
        volume: Union[str, None] = None,
        generic: bool = False,
    ) -> Union[dict, bytes]:
        """Create a new data set profile."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            data_set_request = DataSetRequest(data_set, "set", volume, generic)
            self._build_xml_segments(data_set_request)
            return self._make_request(data_set_request)
        try:
            profile = self.extract(
                data_set, volume=volume, generic=generic, profile_only=True
            )
            if self._get_field(profile, "base", "name") == data_set.lower():
                raise AddOperationError(data_set, self._profile_type)
        except SecurityRequestError as exception:
            if not exception.contains_error_message(self._profile_type, "ICH35003I"):
                raise exception
        self._build_segment_trait_dictionary(traits)
        data_set_request = DataSetRequest(data_set, "set", volume, generic)
        self._build_xml_segments(data_set_request)
        return self._make_request(data_set_request)

    def alter(
        self,
        data_set: str,
        traits: dict,
        volume: Union[str, None] = None,
        generic: bool = False,
    ) -> Union[dict, bytes]:
        """Alter an existing data set profile."""
        if self._generate_requests_only:
            self._build_segment_trait_dictionary(traits)
            data_set_request = DataSetRequest(data_set, "set", volume, generic)
            self._build_xml_segments(data_set_request, alter=True)
            return self._make_request(data_set_request, irrsmo00_precheck=True)
        try:
            profile = self.extract(
                data_set, volume=volume, generic=generic, profile_only=True
            )
        except SecurityRequestError:
            raise AlterOperationError(data_set, self._profile_type)
        if not self._get_field(profile, "base", "name") == data_set.lower():
            raise AlterOperationError(data_set, self._profile_type)
        self._build_segment_trait_dictionary(traits)
        data_set_request = DataSetRequest(data_set, "set", volume, generic)
        self._build_xml_segments(data_set_request, alter=True)
        return self._make_request(data_set_request, irrsmo00_precheck=True)

    def extract(
        self,
        data_set: str,
        segments: List[str] = [],
        volume: Union[str, None] = None,
        generic: bool = False,
        profile_only: bool = False,
    ) -> Union[dict, bytes]:
        """Extract a data set profile."""
        self._build_segment_dictionary(segments)
        data_set_request = DataSetRequest(data_set, "listdata", volume, generic)
        self._build_xml_segments(data_set_request, extract=True)
        result = self._extract_and_check_result(data_set_request)
        if profile_only:
            return self._get_profile(result)
        return result

    def delete(
        self,
        data_set: str,
        volume: Union[str, None] = None,
        generic: bool = False,
    ) -> Union[dict, bytes]:
        """Delete a data set profile."""
        data_set_request = DataSetRequest(data_set, "del", volume, generic)
        return self._make_request(data_set_request)

    # ============================================================================
    # Private/Protected Utility Functions
    # ============================================================================
    def _format_profile(self, result: dict) -> None:
        """Format profile extract data into a dictionary."""
        messages = result["securityResult"]["dataSet"]["commands"][0]["messages"]
        indexes = [
            i
            for i in range(len(messages) - 1)
            if messages[i] and "INFORMATION FOR DATASET " in messages[i]
        ]
        indexes.append(len(messages))
        profiles = []
        for i in range(len(indexes) - 1):
            profile = self._format_profile_generic(
                messages[indexes[i] : indexes[i + 1]]
            )
            # Post processing
            if "(g)" in profile["base"].get("name"):
                profile["base"]["generic"] = True
                profile["base"]["name"] = self._cast_from_str(
                    profile["base"].get("name").split(" ")[0]
                )
            else:
                profile["base"]["generic"] = False

            if profile["base"].get("installationData"):
                profile["base"]["installationData"] = " ".join(
                    profile["base"]["installationData"]
                )
            if profile["base"].get("notify") == [None, "user", "to", "be", "notified"]:
                profile["base"]["notify"] = None
            profiles.append(profile)

        del result["securityResult"]["dataSet"]["commands"][0]["messages"]
        result["securityResult"]["dataSet"]["commands"][0]["profiles"] = profiles
